import torch
from typing import Callable, List, Any, Union, Sequence
from deepchem.utils.differentiation_utils import LinearOperator, \
    get_pure_function, PureFunction
from deepchem.utils.pytorch_utils import TensorNonTensorSeparator


def jac(
    fcn: Callable[..., torch.Tensor],
    params: Sequence[Any],
    idxs: Union[None, int,
                Sequence[int]] = None) -> Union[LinearOperator, List]:
    """
    Returns the LinearOperator that acts as the jacobian of the params.
    The shape of LinearOperator is (nout, nin) where `nout` and `nin` are the
    total number of elements in the output and the input, respectively.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.differentiation_utils import jac
    >>> def fcn(x, y):
    ...     return x * y
    >>> x = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
    >>> y = torch.tensor([4.0, 5.0, 6.0], requires_grad=True)
    >>> jac(fcn, [x, y])
    [LinearOperator (_Jac) with shape (3, 3), dtype = torch.float32, device = cpu, LinearOperator (_Jac) with shape (3, 3), dtype = torch.float32, device = cpu]

    Parameters
    ----------
    fcn: Callable[...,torch.Tensor]
        Callable with tensor output and arbitrary numbers of input parameters.
    params: Sequence[Any]
        List of input parameters of the function.
    idxs: int or list of int or None
        List of the parameters indices to get the jacobian.
        The pointed parameters in `params` must be tensors and requires_grad.
        If it is None, then it will return all jacobian for all parameters that
        are tensor which requires_grad.

    Returns
    -------
    linops: Union[LinearOperator, List]
        List of LinearOperator of the jacobian

    """
    # check idxs
    idxs_list = _setup_idxs(idxs, params)

    # make the function a functional (depends on all parameters in the object)
    pfcn = get_pure_function(fcn)
    res = [_Jac(pfcn, params, idx) for idx in idxs_list]
    if isinstance(idxs, int):
        return res[0]
    return res


class _Jac(LinearOperator):
    """Jacobian of a function with respect to a parameter in the function.

    Examples
    --------
    >>> import torch
    >>> def fcn(x, y):
    ...     return x * y
    >>> x = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
    >>> y = torch.tensor([4.0, 5.0, 6.0], requires_grad=True)
    >>> pfcn = get_pure_function(fcn)
    >>> _Jac(pfcn, [x, y], 1)
    LinearOperator (_Jac) with shape (3, 3), dtype = torch.float32, device = cpu

    """

    def __init__(self,
                 fcn: PureFunction,
                 params: Sequence[Any],
                 idx: int,
                 is_hermitian=False) -> None:
        """Initialize the _Jac object.

        Parameters
        ----------
        fcn: PureFunction
            The function that will be differentiated.
        params: Sequence[Any]
            List of input parameters of the function.
        idx: int
            The index of the parameter to be differentiated.
        is_hermitian: bool
            If True, then the LinearOperator is hermitian.

        """
        yparam = params[idx]
        with torch.enable_grad():
            yout = fcn(*params)  # (*nout)
            v = torch.ones_like(yout).to(
                yout.device).requires_grad_()  # (*nout)
            dfdy, = torch.autograd.grad(yout, (yparam,),
                                        grad_outputs=v,
                                        create_graph=True)  # (*nin)

        inshape = yparam.shape
        outshape = yout.shape
        nin = torch.numel(yparam)
        nout = torch.numel(yout)

        super(_Jac, self).__init__(shape=(nout, nin),
                                   is_hermitian=is_hermitian,
                                   dtype=yparam.dtype,
                                   device=yparam.device)

        self.fcn = fcn
        self.yparam = yparam
        self.params = list(params)
        self.objparams = fcn.objparams()
        self.yout = yout
        self.v = v
        self.idx = idx
        self.dfdy = dfdy
        self.inshape = inshape
        self.outshape = outshape
        self.nin = nin
        self.nout = nout

        # params tensor is the LinearOperator's parameters
        self.param_sep = TensorNonTensorSeparator(params)
        self.params_tensor = self.param_sep.get_tensor_params()
        self.id_params_tensor = [id(param) for param in self.params_tensor]
        self.id_objparams_tensor = [id(param) for param in self.objparams]

    def _getparamnames(self, prefix: str = "") -> List[str]:
        """Returns the parameter names of the LinearOperator.

        Parameters
        ----------
        prefix: str
            Prefix of the parameter names.

        Returns
        -------
        paramnames: List[str]
            List of parameter names.

        """
        return [prefix + "yparam"] + \
               [prefix + ("params_tensor[%d]" % i) for i in range(len(self.params_tensor))] + \
               [prefix + ("objparams[%d]" % i) for i in range(len(self.objparams))]

    def _mv(self, gy: torch.Tensor) -> torch.Tensor:
        """Matrix-vector multiplication.

        Parameters
        ----------
        x: torch.Tensor
            The vector with shape ``(...,q)`` where the linear operation is
            operated on.

        Returns
        -------
        torch.Tensor
            The result of the linear operation with shape ``(...,p)``

        """

        # if the object parameter is still the same, then use the pre-calculated values
        if self.__param_tensors_unchanged():
            v = self.v
            dfdy = self.dfdy
        # otherwise, reevaluate by replacing the parameters with the new tensor params
        else:
            with torch.enable_grad(), self.fcn.useobjparams(self.objparams):
                self.__update_params()
                yparam = self.params[self.idx]
                yout = self.fcn(*self.params)  # (*nout)
                v = torch.ones_like(yout).to(
                    yout.device).requires_grad_()  # (*nout)
                dfdy, = torch.autograd.grad(yout, (yparam,),
                                            grad_outputs=v,
                                            create_graph=True)  # (*nin)

        gy1 = gy.reshape(-1, self.nin)  # (nbatch, nin)
        nbatch = gy1.shape[0]
        dfdyfs_list = []
        for i in range(nbatch):
            dfdyf, = torch.autograd.grad(
                dfdy, (v,),
                grad_outputs=gy1[i].reshape(self.inshape),
                retain_graph=True,
                create_graph=torch.is_grad_enabled())  # (*nout)
            dfdyfs_list.append(dfdyf.unsqueeze(0))
        dfdyfs = torch.cat(dfdyfs_list, dim=0)  # (nbatch, *nout)

        res = dfdyfs.reshape(*gy.shape[:-1], self.nout)  # (..., nout)
        res = connect_graph(res, self.params_tensor)
        res = connect_graph(res, self.objparams)
        return res

    def _rmv(self, gout: torch.Tensor) -> torch.Tensor:
        """Matrix-vector adjoint multiplication.

        Parameters
        ----------
        x: torch.Tensor
            The vector with shape ``(...,p)`` where the adjoint linear operation is
            operated on.

        Returns
        -------
        torch.Tensor
            The result of the adjoint linear operation with shape ``(...,q)``

        """
        # self.yfcn: (*nin)
        if self.__param_tensors_unchanged():
            yout = self.yout
            yparam = self.yparam
        else:
            with torch.enable_grad(), self.fcn.useobjparams(self.objparams):
                self.__update_params()
                yparam = self.params[self.idx]
                yout = self.fcn(*self.params)  # (*nout)

        gout1 = gout.reshape(-1, self.nout)  # (nbatch, nout)
        nbatch = gout1.shape[0]
        dfdy_list = []
        for i in range(nbatch):
            one_dfdy, = torch.autograd.grad(
                yout, (yparam,),
                grad_outputs=gout1[i].reshape(self.outshape),
                retain_graph=True,
                create_graph=torch.is_grad_enabled())  # (*nin)
            dfdy_list.append(one_dfdy.unsqueeze(0))
        dfdy = torch.cat(dfdy_list, dim=0)  # (nbatch, *nin)

        res = dfdy.reshape(*gout.shape[:-1], self.nin)  # (..., nin)
        res = connect_graph(res, self.params_tensor)
        res = connect_graph(res, self.objparams)
        return res  # (..., nin)

    def __param_tensors_unchanged(self):
        """Check if the parameters are the same as the last time.
        If they are the same, then we can reuse the pre-calculated values.

        Returns
        -------
        bool
            True if the parameters are the same as the last time.

        """
        return [id(param) for param in self.params_tensor] == self.id_params_tensor and \
               [id(param) for param in self.objparams] == self.id_objparams_tensor

    def __update_params(self):
        self.params = self.param_sep.reconstruct_params(self.params_tensor)


def connect_graph(out: torch.Tensor, params: Sequence[Any]):
    """Just to have a dummy graph, in case there is a parameter that
    is disconnected in calculating df/dy.

    Examples
    --------
    >>> import torch
    >>> x = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
    >>> y = torch.tensor([4.0, 5.0, 6.0], requires_grad=True)
    >>> out = x * y
    >>> connect_graph(out, [x, y])
    tensor([ 4., 10., 18.], grad_fn=<AddBackward0>)

    Parameters
    ----------
    out: torch.Tensor
        The output tensor. It will be added with a dummy graph.
    params: Sequence[Any]
        List of parameters that will be added with a dummy graph.

    Returns
    -------
    out: torch.Tensor
        The output tensor with a dummy graph.

    """
    return out + sum([p.reshape(-1)[0] * 0 for p in params])


def _setup_idxs(idxs: Union[None, int, Sequence[int]],
                params: Sequence[Any]) -> Sequence[int]:
    """Check the idxs and return the list of indices.

    Examples
    --------
    >>> import torch
    >>> import numpy as np
    >>> x = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
    >>> y = torch.tensor([4.0, 5.0, 6.0], requires_grad=True)
    >>> _setup_idxs(None, [x, y])
    [0, 1]

    Parameters
    ----------
    idxs: int or list of int or None
        List of the parameters indices to get the jacobian.
        The pointed parameters in `params` must be tensors and requires_grad.
        If it is None, then it will return all jacobian for all parameters that
        are tensor which requires_grad.
    params: Sequence[Any]
        List of input parameters of the function.

    Returns
    -------
    idxs: list of int
        List of the parameters indices to get the jacobian.

    """
    if idxs is None:
        idxs = [
            i for i, t in enumerate(params)
            if isinstance(t, torch.Tensor) and t.requires_grad
        ]
    elif isinstance(idxs, int):
        idxs = [idxs]

    for p in idxs:
        assert isinstance(params[p], torch.Tensor) and params[
            p].requires_grad, "The %d-th element (0-based) must be a tensor which requires grad" % p
    return idxs



"""
Cell Counting Dataset.

Loads the cell counting dataset from
http://www.robots.ox.ac.uk/~vgg/research/counting/index_org.html. Labels aren't
available for this dataset, so only raw images are provided.
"""
import os
import deepchem as dc
from deepchem.molnet.load_function.molnet_loader import TransformerGenerator, _MolnetLoader
from deepchem.data import Dataset
from typing import List, Optional, Tuple, Union

CELL_COUNTING_URL = 'http://www.robots.ox.ac.uk/~vgg/research/counting/cells.zip'
CELL_COUNTING_TASKS: List[str] = []


class _CellCountingLoader(_MolnetLoader):

    def create_dataset(self) -> Dataset:
        dataset_file = os.path.join(self.data_dir, "cells.zip")
        if not os.path.exists(dataset_file):
            dc.utils.data_utils.download_url(url=CELL_COUNTING_URL,
                                             dest_dir=self.data_dir)
        loader = dc.data.ImageLoader()
        return loader.featurize(dataset_file)


def load_cell_counting(
    splitter: Union[dc.splits.Splitter, str, None] = None,
    transformers: List[Union[TransformerGenerator, str]] = [],
    reload: bool = True,
    data_dir: Optional[str] = None,
    save_dir: Optional[str] = None,
    **kwargs
) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
    """Load Cell Counting dataset.

    Loads the cell counting dataset from http://www.robots.ox.ac.uk/~vgg/research/counting/index_org.html.

    Parameters
    ----------
    splitter: Splitter or str
        the splitter to use for splitting the data into training, validation, and
        test sets.  Alternatively you can pass one of the names from
        dc.molnet.splitters as a shortcut.  If this is None, all the data
        will be included in a single dataset.
    transformers: list of TransformerGenerators or strings
        the Transformers to apply to the data.  Each one is specified by a
        TransformerGenerator or, as a shortcut, one of the names from
        dc.molnet.transformers.
    reload: bool
        if True, the first call for a particular featurizer and splitter will cache
        the datasets to disk, and subsequent calls will reload the cached datasets.
    data_dir: str
        a directory to save the raw data in
    save_dir: str
        a directory to save the dataset in
    """
    featurizer = dc.feat.UserDefinedFeaturizer([])  # Not actually used
    loader = _CellCountingLoader(featurizer, splitter, transformers,
                                 CELL_COUNTING_TASKS, data_dir, save_dir,
                                 **kwargs)
    return loader.load_dataset('cell_counting', reload)



"""
bace dataset loader.
"""
import os
import deepchem as dc
from deepchem.molnet.load_function.molnet_loader import TransformerGenerator, _MolnetLoader
from deepchem.data import Dataset
from typing import List, Optional, Tuple, Union

BACE_URL = "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/bace.csv"
BACE_REGRESSION_TASKS = ["pIC50"]
BACE_CLASSIFICATION_TASKS = ["Class"]


class _BaceLoader(_MolnetLoader):

    def create_dataset(self) -> Dataset:
        dataset_file = os.path.join(self.data_dir, "bace.csv")
        if not os.path.exists(dataset_file):
            dc.utils.data_utils.download_url(url=BACE_URL,
                                             dest_dir=self.data_dir)
        loader = dc.data.CSVLoader(tasks=self.tasks,
                                   feature_field="mol",
                                   featurizer=self.featurizer)
        return loader.create_dataset(dataset_file, shard_size=8192)


def load_bace_regression(
    featurizer: Union[dc.feat.Featurizer, str] = 'ECFP',
    splitter: Union[dc.splits.Splitter, str, None] = 'scaffold',
    transformers: List[Union[TransformerGenerator, str]] = ['normalization'],
    reload: bool = True,
    data_dir: Optional[str] = None,
    save_dir: Optional[str] = None,
    **kwargs
) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
    """ Load BACE dataset, regression labels

    The BACE dataset provides quantitative IC50 and qualitative (binary label)
    binding results for a set of inhibitors of human beta-secretase 1 (BACE-1).

    All data are experimental values reported in scientific literature over the
    past decade, some with detailed crystal structures available. A collection
    of 1522 compounds is provided, along with the regression labels of IC50.

    Scaffold splitting is recommended for this dataset.

    The raw data csv file contains columns below:

    - "mol" - SMILES representation of the molecular structure
    - "pIC50" - Negative log of the IC50 binding affinity
    - "class" - Binary labels for inhibitor

    Parameters
    ----------
    featurizer: Featurizer or str
        the featurizer to use for processing the data.  Alternatively you can pass
        one of the names from dc.molnet.featurizers as a shortcut.
    splitter: Splitter or str
        the splitter to use for splitting the data into training, validation, and
        test sets.  Alternatively you can pass one of the names from
        dc.molnet.splitters as a shortcut.  If this is None, all the data
        will be included in a single dataset.
    transformers: list of TransformerGenerators or strings
        the Transformers to apply to the data.  Each one is specified by a
        TransformerGenerator or, as a shortcut, one of the names from
        dc.molnet.transformers.
    reload: bool
        if True, the first call for a particular featurizer and splitter will cache
        the datasets to disk, and subsequent calls will reload the cached datasets.
    data_dir: str
        a directory to save the raw data in
    save_dir: str
        a directory to save the dataset in

    References
    ----------
    .. [1] Subramanian, Govindan, et al. "Computational modeling of β-secretase 1
        (BACE-1) inhibitors using ligand based approaches." Journal of chemical
        information and modeling 56.10 (2016): 1936-1949.
    """
    loader = _BaceLoader(featurizer, splitter, transformers,
                         BACE_REGRESSION_TASKS, data_dir, save_dir, **kwargs)
    return loader.load_dataset('bace_r', reload)


def load_bace_classification(
    featurizer: Union[dc.feat.Featurizer, str] = 'ECFP',
    splitter: Union[dc.splits.Splitter, str, None] = 'scaffold',
    transformers: List[Union[TransformerGenerator, str]] = ['balancing'],
    reload: bool = True,
    data_dir: Optional[str] = None,
    save_dir: Optional[str] = None,
    **kwargs
) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
    """ Load BACE dataset, classification labels

    BACE dataset with classification labels ("class").

    Parameters
    ----------
    featurizer: Featurizer or str
        the featurizer to use for processing the data.  Alternatively you can pass
        one of the names from dc.molnet.featurizers as a shortcut.
    splitter: Splitter or str
        the splitter to use for splitting the data into training, validation, and
        test sets.  Alternatively you can pass one of the names from
        dc.molnet.splitters as a shortcut.  If this is None, all the data
        will be included in a single dataset.
    transformers: list of TransformerGenerators or strings
        the Transformers to apply to the data.  Each one is specified by a
        TransformerGenerator or, as a shortcut, one of the names from
        dc.molnet.transformers.
    reload: bool
        if True, the first call for a particular featurizer and splitter will cache
        the datasets to disk, and subsequent calls will reload the cached datasets.
    data_dir: str
        a directory to save the raw data in
    save_dir: str
        a directory to save the dataset in
    """
    loader = _BaceLoader(featurizer, splitter, transformers,
                         BACE_CLASSIFICATION_TASKS, data_dir, save_dir,
                         **kwargs)
    return loader.load_dataset('bace_c', reload)



"""Generative Adversarial Networks."""
import numpy as np
try:
    import torch
    import torch.nn as nn
except ModuleNotFoundError:
    raise ImportError('These classes require PyTorch to be installed.')
import time
from typing import Callable, Any, List, Optional, Tuple, Union
from deepchem.models.torch_models.torch_model import TorchModel


class GAN(nn.Module):
    """Builder class for Generative Adversarial Networks.

    A Generative Adversarial Network (GAN) is a type of generative model. It
    consists of two parts called the "generator" and the "discriminator". The
    generator takes random noise as input and transforms it into an output that
    (hopefully) resembles the training data. The discriminator takes a set of
    samples as input and tries to distinguish the real training samples from the
    ones created by the generator. Both of them are trained together. The
    discriminator tries to get better and better at telling real from false data,
    while the generator tries to get better and better at fooling the discriminator.

    Examples
    --------
    Importing necessary modules

    >>> import deepchem as dc
    >>> from deepchem.models.torch_models.gan import GAN
    >>> import torch
    >>> import torch.nn as nn
    >>> import torch.nn.functional as F

    Creating a Generator

    >>> class Generator(nn.Module):
    ...     def __init__(self, noise_input_shape, conditional_input_shape):
    ...         super(Generator, self).__init__()
    ...         self.noise_input_shape = noise_input_shape
    ...         self.conditional_input_shape = conditional_input_shape
    ...         self.noise_dim = noise_input_shape[1:]
    ...         self.conditional_dim = conditional_input_shape[1:]
    ...         input_dim = sum(self.noise_dim) + sum(self.conditional_dim)
    ...         self.output = nn.Linear(input_dim, 1)
    ...     def forward(self, input):
    ...         noise_input, conditional_input = input
    ...         inputs = torch.cat((noise_input, conditional_input), dim=1)
    ...         output = self.output(inputs)
    ...         return output

    Creating a Discriminator

    >>> class Discriminator(nn.Module):
    ...     def __init__(self, data_input_shape, conditional_input_shape):
    ...         super(Discriminator, self).__init__()
    ...         self.data_input_shape = data_input_shape
    ...         self.conditional_input_shape = conditional_input_shape
    ...         # Extracting the actual data dimension
    ...         data_dim = data_input_shape[1:]
    ...         # Extracting the actual conditional dimension
    ...         conditional_dim = conditional_input_shape[1:]
    ...         input_dim = sum(data_dim) + sum(conditional_dim)
    ...         # Define the dense layers
    ...         self.dense1 = nn.Linear(input_dim, 10)
    ...         self.dense2 = nn.Linear(10, 1)
    ...     def forward(self, input):
    ...         data_input, conditional_input = input
    ...         # Concatenate data_input and conditional_input along the second dimension
    ...         discrim_in = torch.cat((data_input, conditional_input), dim=1)
    ...         # Pass the concatenated input through the dense layers
    ...         x = F.relu(self.dense1(discrim_in))
    ...         output = torch.sigmoid(self.dense2(x))
    ...         return output

    Defining an Example GAN class

    >>> class ExampleGAN(dc.models.torch_models.GAN):
    ...    def get_noise_input_shape(self):
    ...        return (16,2,)
    ...    def get_data_input_shapes(self):
    ...        return [(16,1,)]
    ...    def get_conditional_input_shapes(self):
    ...        return [(16,1,)]
    ...    def create_generator(self):
    ...        noise_dim = self.get_noise_input_shape()
    ...        conditional_dim = self.get_conditional_input_shapes()[0]
    ...        return nn.Sequential(Generator(noise_dim, conditional_dim))
    ...    def create_discriminator(self):
    ...        data_input_shape = self.get_data_input_shapes()[0]
    ...        conditional_input_shape = self.get_conditional_input_shapes()[0]
    ...        return nn.Sequential(
    ...            Discriminator(data_input_shape, conditional_input_shape))

    Defining the GAN

    >>> batch_size = 16
    >>> noise_shape = (batch_size, 2,)
    >>> data_shape = [(batch_size, 1,)]
    >>> conditional_shape = [(batch_size, 1,)]
    >>> def create_generator(noise_dim, conditional_dim):
    ...     noise_dim = noise_dim
    ...     conditional_dim = conditional_dim[0]
    ...     return nn.Sequential(Generator(noise_dim, conditional_dim))
    >>> def create_discriminator(data_input_shape, conditional_input_shape):
    ...     data_input_shape = data_input_shape[0]
    ...     conditional_input_shape = conditional_input_shape[0]
    ...     return nn.Sequential(
    ...         Discriminator(data_input_shape, conditional_input_shape))
    >>> gan = ExampleGAN(noise_shape,
    ...              data_shape,
    ...              conditional_shape,
    ...              create_generator(noise_shape, conditional_shape),
    ...              create_discriminator(data_shape, conditional_shape),
    ...              device='cpu')
    >>> noise = torch.rand(*gan.noise_input_shape)
    >>> real_data = torch.rand(*gan.data_input_shape[0])
    >>> conditional = torch.rand(*gan.conditional_input_shape[0])
    >>> gen_loss, disc_loss = gan([noise, real_data, conditional])

    References
    ----------
    .. [1] Goodfellow, Ian, et al. "Generative adversarial nets." Advances in
         neural information processing systems. 2014.

    .. [2] Arora et al., “Generalization and Equilibrium in Generative
         Adversarial Nets (GANs)” (https://arxiv.org/abs/1703.00573)
    """

    def __init__(self,
                 noise_input_shape: tuple,
                 data_input_shape: list,
                 conditional_input_shape: list,
                 generator_fn: Callable,
                 discriminator_fn: Callable,
                 device: torch.device,
                 n_generators: int = 1,
                 n_discriminators: int = 1,
                 create_discriminator_loss: Optional[Callable] = None,
                 create_generator_loss: Optional[Callable] = None,
                 _call_discriminator: Optional[Callable] = None,
                 **kwargs):
        """Construct a GAN.

        In addition to the parameters listed below, this class accepts all the
        keyword arguments from KerasModel.

        Parameters
        ----------
        noise_input_shape: tuple
            the shape of the noise input to the generator.  The first dimension
            (corresponding to the batch size) should be omitted.
        data_input_shape: list of tuple
            the shapes of the inputs to the discriminator.  The first dimension
            (corresponding to the batch size) should be omitted.
        conditional_input_shape: list of tuple
            the shapes of the conditional inputs to the generator and discriminator.
            The first dimension (corresponding to the batch size) should be omitted.
            If there are no conditional inputs, this should be an empty list.
        generator_fn: Callable
            a function that returns a generator.  It will be called with no arguments.
            The returned value should be a nn.Module whose input is a list
            containing a batch of noise, followed by any conditional inputs.  The
            number and shapes of its outputs must match the return value from
            get_data_input_shapes(), since generated data must have the same form as
            training data.
        discriminator_fn: Callable
            a function that returns a discriminator.  It will be called with no
            arguments.  The returned value should be a nn.Module whose input is a
            list containing a batch of data, followed by any conditional inputs.  Its
            output should be a one dimensional tensor containing the probability of
            each sample being a training sample.
        device: torch.device
            the device to use for training
        n_generators: int
            the number of generators to include
        n_discriminators: int
            the number of discriminators to include
        create_discriminator_loss: Callable
            a function that returns the loss function for the discriminator.  It will
            be called with two arguments: the output from the discriminator on a
            batch of training data, and the output from the discriminator on a batch
            of generated data.  The default implementation is appropriate for most
            cases.  Subclasses can override this if the need to customize it.
        create_generator_loss: Callable
            a function that returns the loss function for the generator.  It will be
            called with one argument: the output from the discriminator on a batch of
            generated data.  The default implementation is appropriate for most
            cases.  Subclasses can override this if the need to customize it.
        _call_discriminator: Callable
            a function that invokes the discriminator on a set of inputs.  It will be
            called with three arguments: the discriminator to invoke, the list of
            data inputs, and the list of conditional inputs.  The default
            implementation is appropriate for most cases.  Subclasses can override
            this if the need to customize it.
        """
        super(GAN, self).__init__(**kwargs)
        self.n_generators = n_generators
        self.n_discriminators = n_discriminators
        self.noise_input_shape = noise_input_shape
        self.data_input_shape = data_input_shape
        self.conditional_input_shape = conditional_input_shape
        self.create_generator = generator_fn
        self.create_discriminator = discriminator_fn
        self.device = device
        if create_discriminator_loss is not None:
            self.create_discriminator_loss = create_discriminator_loss  # type: ignore
        if create_generator_loss is not None:
            self.create_generator_loss = create_generator_loss  # type: ignore
        if _call_discriminator is not None:
            self._call_discriminator = _call_discriminator  # type: ignore

        # Inputs
        # Noise Input
        self.noise_input = nn.Parameter(torch.empty(self.noise_input_shape))
        # Data Input
        self.data_input_list = [
            nn.Parameter(torch.empty(s)) for s in self.data_input_shape
        ]
        # Data Inputs
        self.data_inputs = [
            "data_input_%d" % i for i in range(len(self.data_input_list))
        ]

        # Conditional Input
        self.conditional_input_list = [
            nn.Parameter(torch.empty(s)) for s in self.conditional_input_shape
        ]
        # Conditional Inputs
        self.conditional_inputs = [
            "conditional_input_%d" % i
            for i in range(len(self.conditional_input_list))
        ]

        self.data_input_layers = []
        for idx, shape in enumerate(self.data_input_shape):
            self.data_input_layers.append(nn.Parameter(torch.empty(shape)))
        self.conditional_input_layers = []
        for idx, shape in enumerate(self.conditional_input_shape):
            self.conditional_input_layers.append(
                nn.Parameter(torch.empty(shape)))

        # Generators
        self.generators = nn.ModuleList()
        self.gen_variables = nn.ParameterList()
        for i in range(n_generators):
            generator = self.create_generator()
            self.generators.append(generator)
            self.gen_variables += list(generator.parameters())

        # Discriminators
        self.discriminators = nn.ModuleList()
        self.discrim_variables = nn.ParameterList()

        for i in range(n_discriminators):
            discriminator = self.create_discriminator()
            self.discriminators.append(discriminator)
            self.discrim_variables += list(discriminator.parameters())

    def forward(self, inputs) -> Tuple[torch.Tensor, torch.Tensor]:
        """Compute the output of the GAN.

        Parameters
        ----------
        inputs: list of Tensor
            the inputs to the GAN. The first element must be a batch of noise,
            followed by data inputs and any conditional inputs.

        Returns
        -------
        total_gen_loss: Tensor
            the total loss for the generator
        total_discrim_loss: Tensor
            the total loss for the discriminator
        """

        n_generators = self.n_generators
        n_discriminators = self.n_discriminators
        noise_input, data_input_layers, conditional_input_layers = inputs[
            0], inputs[1], inputs[2]

        self.noise_input.data = noise_input
        self.conditional_input_layers = [conditional_input_layers]
        self.data_input_layers = [data_input_layers]

        # Forward pass through generators
        generator_outputs = [
            gen(_list_or_tensor([[noise_input] + self.conditional_input_layers
                                ])) for gen in self.generators
        ]

        # Forward pass through discriminators
        discrim_train_outputs = [
            self._call_discriminator(disc, self.data_input_layers,
                                     self.conditional_input_layers, True)
            for disc in self.discriminators
        ]

        discrim_gen_outputs = [
            self._call_discriminator(disc, [gen_output],
                                     self.conditional_input_layers, False)
            for disc in self.discriminators
            for gen_output in generator_outputs
        ]

        # Compute loss functions
        gen_losses = [
            self.create_generator_loss(d) for d in discrim_gen_outputs
        ]
        discrim_losses = [
            self.create_discriminator_loss(
                discrim_train_outputs[i],
                discrim_gen_outputs[i * self.n_generators + j])
            for i in range(self.n_discriminators)
            for j in range(self.n_generators)
        ]

        # Compute the weighted errors
        if n_generators == 1 and n_discriminators == 1:
            total_gen_loss = gen_losses[0]
            total_discrim_loss = discrim_losses[0]
        else:
            # Create learnable weights for the generators and discriminators.

            gen_alpha = nn.Parameter(torch.ones(1, n_generators))
            gen_weights = nn.Parameter(torch.softmax(gen_alpha, dim=1))

            discrim_alpha = nn.Parameter(torch.ones(1, n_discriminators))
            discrim_weights = nn.Parameter(torch.softmax(discrim_alpha, dim=1))

            # Compute the weighted errors

            discrim_weights_n = discrim_weights.view(-1, self.n_discriminators,
                                                     1)
            gen_weights_n = gen_weights.view(-1, 1, self.n_generators)

            weight_products = torch.mul(discrim_weights_n, gen_weights_n)
            weight_products = weight_products.view(
                -1, self.n_generators * self.n_discriminators)
            weight_products = weight_products.to(self.device)

            stacked_gen_loss = torch.stack(gen_losses, dim=0).to(self.device)
            stacked_discrim_loss = torch.stack(discrim_losses,
                                               dim=0).to(self.device)

            total_gen_loss = torch.sum(stacked_gen_loss * weight_products)
            total_discrim_loss = torch.sum(stacked_discrim_loss *
                                           weight_products)

            self.gen_variables += [gen_alpha]
            self.discrim_variables += [gen_alpha]
            self.discrim_variables += [discrim_alpha]

            # Add an entropy term to the loss.

            entropy = -(
                torch.sum(torch.log(gen_weights)) / n_generators +
                torch.sum(torch.log(discrim_weights)) / n_discriminators)
            total_discrim_loss = total_discrim_loss + entropy

        return total_gen_loss, total_discrim_loss

    def _call_discriminator(self, discriminator: nn.Module, inputs: List,
                            conditional_input_layers: List, train: bool):
        """Invoke the discriminator on a set of inputs.

        This is a separate method so WGAN can override it and also return the
        gradient penalty.

        Parameters
        ----------
        discriminator: nn.Module
            the discriminator to invoke
        inputs: list of Tensor
            the inputs to the discriminator.  The first element must be a batch of
            data, followed by any conditional inputs.
        conditional_input_layers: list of Tensor
            the conditional inputs to the discriminator
        train: bool
            if True, the discriminator should be invoked in training mode. If False,
            it should be invoked in inference mode.

        Returns
        -------
        the output from the discriminator
        """
        return discriminator(_list_or_tensor(inputs + conditional_input_layers))

    def get_noise_batch(self, batch_size: int) -> np.ndarray:
        """Get a batch of random noise to pass to the generator.

        This should return a NumPy array whose shape matches the one returned by
        get_noise_input_shape().  The default implementation returns normally
        distributed values.  Subclasses can override this to implement a different
        distribution.

        Parameters
        ----------
        batch_size: int
            the number of samples to generate

        Returns
        -------
        random_noise: ndarray
            a batch of random noise
        """
        size = list(self.noise_input_shape)
        size = [batch_size] + size[1:]
        return np.random.normal(size=size)

    def create_generator_loss(self,
                              discrim_output: torch.Tensor) -> torch.Tensor:
        """Create the loss function for the generator.

        The default implementation is appropriate for most cases.  Subclasses can
        override this if the need to customize it.

        Parameters
        ----------
        discrim_output: Tensor
            the output from the discriminator on a batch of generated data.  This is
            its estimate of the probability that each sample is training data.

        Returns
        -------
        output: Tensor
            A Tensor equal to the loss function to use for optimizing the generator.
        """
        return -torch.mean(torch.log(discrim_output + 1e-10))

    def create_discriminator_loss(
            self, discrim_output_train: torch.Tensor,
            discrim_output_gen: torch.Tensor) -> torch.Tensor:
        """Create the loss function for the discriminator.

        The default implementation is appropriate for most cases.  Subclasses can
        override this if the need to customize it.

        Parameters
        ----------
        discrim_output_train: Tensor
            the output from the discriminator on a batch of training data.  This is
            its estimate of the probability that each sample is training data.
        discrim_output_gen: Tensor
            the output from the discriminator on a batch of generated data.  This is
            its estimate of the probability that each sample is training data.

        Returns
        -------
        output: Tensor
            A Tensor equal to the loss function to use for optimizing the discriminator.
        """
        return -torch.mean(
            torch.log(discrim_output_train + 1e-10) +
            torch.log(1 - discrim_output_gen + 1e-10))

    def discrim_loss_fn(self, outputs: List, labels: List[torch.Tensor],
                        weights: List[torch.Tensor]) -> Any:
        """Function to get the discriminator loss from the fit_generator output

        Parameters
        ----------
        outputs: list of Tensor
            the output from the discriminator on a batch of training data.  This is
            its estimate of the probability that each sample is training data.
        labels: Tensor
            the labels for the batch.  These are ignored.
        weights: Tensor
            the weights for the batch.  These are ignored.

        Returns
        -------
        the value of the discriminator loss from the fit_generator output.
        """
        discrim_output_train, discrim_output_gen = outputs
        return discrim_output_gen

    def gen_loss_fn(self, outputs: List, labels: List[torch.Tensor],
                    weights: List[torch.Tensor]) -> torch.Tensor:
        """Function to get the Generator loss from the fit_generator output

        Parameters
        ----------
        outputs: Tensor
            the output from the discriminator on a batch of generated data.  This is
            its estimate of the probability that each sample is training data.
        labels: Tensor
            the labels for the batch.  These are ignored.
        weights: Tensor
            the weights for the batch.  These are ignored.

        Returns
        -------
        the value of the generator loss function for this input.
        """
        discrim_output_train, discrim_output_gen = outputs
        return discrim_output_train


def _list_or_tensor(
    inputs: Union[list,
                  torch.Tensor]) -> Union[torch.Tensor, List[torch.Tensor]]:
    """Function to convert a list of tensors to a single tensor if possible.

    Parameters
    ----------
    inputs : Union[list, torch.Tensor]
        A list of tensors or a single tensor.

    Returns
    -------
    Union[torch.Tensor, list[torch.Tensor]]
        A single tensor if the input is a list of length 1, otherwise the
        original list.
    """
    if len(inputs) == 1:
        return inputs[0]
    return inputs


class GANModel(TorchModel):
    """Implements Generative Adversarial Networks.

    A Generative Adversarial Network (GAN) is a type of generative model.  It
    consists of two parts called the "generator" and the "discriminator".  The
    generator takes random noise as input and transforms it into an output that
    (hopefully) resembles the training data.  The discriminator takes a set of
    samples as input and tries to distinguish the real training samples from the
    ones created by the generator.  Both of them are trained together.  The
    discriminator tries to get better and better at telling real from false data,
    while the generator tries to get better and better at fooling the discriminator.

    In many cases there also are additional inputs to the generator and
    discriminator.  In that case it is known as a Conditional GAN (CGAN), since it
    learns a distribution that is conditional on the values of those inputs.  They
    are referred to as "conditional inputs".

    Many variations on this idea have been proposed, and new varieties of GANs are
    constantly being proposed.  This class tries to make it very easy to implement
    straightforward GANs of the most conventional types.  At the same time, it
    tries to be flexible enough that it can be used to implement many (but
    certainly not all) variations on the concept.

    To define a GAN, you must create a subclass that provides implementations of
    the following methods:

    get_noise_input_shape()
    get_data_input_shapes()
    create_generator()
    create_discriminator()

    If you want your GAN to have any conditional inputs you must also implement:

    get_conditional_input_shapes()

    The following methods have default implementations that are suitable for most
    conventional GANs.  You can override them if you want to customize their
    behavior:

    create_generator_loss()
    create_discriminator_loss()
    get_noise_batch()

    This class allows a GAN to have multiple generators and discriminators, a model
    known as MIX+GAN.  It is described in [2]
    This can lead to better models, and is especially useful for reducing mode
    collapse, since different generators can learn different parts of the
    distribution.  To use this technique, simply specify the number of generators
    and discriminators when calling the constructor.  You can then tell
    predict_gan_generator() which generator to use for predicting samples.

    Examples
    --------
    Importing necessary modules

    >>> import deepchem as dc
    >>> from deepchem.models.torch_models.gan import GAN
    >>> import torch
    >>> import torch.nn as nn
    >>> import torch.nn.functional as F

    Creating a Generator

    >>> class Generator(nn.Module):
    ...     def __init__(self, noise_input_shape, conditional_input_shape):
    ...         super(Generator, self).__init__()
    ...         self.noise_input_shape = noise_input_shape
    ...         self.conditional_input_shape = conditional_input_shape
    ...         self.noise_dim = noise_input_shape[1:]
    ...         self.conditional_dim = conditional_input_shape[1:]
    ...         input_dim = sum(self.noise_dim) + sum(self.conditional_dim)
    ...         self.output = nn.Linear(input_dim, 1)
    ...     def forward(self, input):
    ...         noise_input, conditional_input = input
    ...         inputs = torch.cat((noise_input, conditional_input), dim=1)
    ...         output = self.output(inputs)
    ...         return output

    Creating a Discriminator

    >>> class Discriminator(nn.Module):
    ...     def __init__(self, data_input_shape, conditional_input_shape):
    ...         super(Discriminator, self).__init__()
    ...         self.data_input_shape = data_input_shape
    ...         self.conditional_input_shape = conditional_input_shape
    ...         # Extracting the actual data dimension
    ...         data_dim = data_input_shape[1:]
    ...         # Extracting the actual conditional dimension
    ...         conditional_dim = conditional_input_shape[1:]
    ...         input_dim = sum(data_dim) + sum(conditional_dim)
    ...         # Define the dense layers
    ...         self.dense1 = nn.Linear(input_dim, 10)
    ...         self.dense2 = nn.Linear(10, 1)
    ...     def forward(self, input):
    ...         data_input, conditional_input = input
    ...         # Concatenate data_input and conditional_input along the second dimension
    ...         discrim_in = torch.cat((data_input, conditional_input), dim=1)
    ...         # Pass the concatenated input through the dense layers
    ...         x = F.relu(self.dense1(discrim_in))
    ...         output = torch.sigmoid(self.dense2(x))
    ...         return output

    Defining an Example GAN class

    >>> class ExampleGANModel(dc.models.torch_models.GANModel):
    ...    def get_noise_input_shape(self):
    ...        return (100,2,)
    ...    def get_data_input_shapes(self):
    ...        return [(100,1,)]
    ...    def get_conditional_input_shapes(self):
    ...        return [(100,1,)]
    ...    def create_generator(self):
    ...        noise_dim = self.get_noise_input_shape()
    ...        conditional_dim = self.get_conditional_input_shapes()[0]
    ...        return nn.Sequential(Generator(noise_dim, conditional_dim))
    ...    def create_discriminator(self):
    ...        data_input_shape = self.get_data_input_shapes()[0]
    ...        conditional_input_shape = self.get_conditional_input_shapes()[0]
    ...        return nn.Sequential(
    ...            Discriminator(data_input_shape, conditional_input_shape))

    Defining a function to generate data

    >>> def generate_batch(batch_size):
    ...     means = 10 * np.random.random([batch_size, 1])
    ...     values = np.random.normal(means, scale=2.0)
    ...     return means, values

    >>> def generate_data(gan, batches, batch_size):
    ...     for _ in range(batches):
    ...         means, values = generate_batch(batch_size)
    ...         batch = {
    ...             gan.data_inputs[0]: values,
    ...             gan.conditional_inputs[0]: means
    ...         }
    ...         yield batch

    Defining the GANModel

    >>> batch_size = 100
    >>> noise_shape = (batch_size, 2,)
    >>> data_shape = [(batch_size, 1,)]
    >>> conditional_shape = [(batch_size, 1,)]
    >>> gan = ExampleGANModel(learning_rate=0.01)
    >>> data = generate_data(gan, 500, 100)
    >>> gan.fit_gan(data, generator_steps=0.5, checkpoint_interval=0)
    >>> means = 10 * np.random.random([1000, 1])
    >>> values = gan.predict_gan_generator(conditional_inputs=[means])

    References
    ----------
    .. [1] Goodfellow, Ian, et al. "Generative adversarial nets." Advances in
         neural information processing systems. 2014.

    .. [2] Arora et al., “Generalization and Equilibrium in Generative
         Adversarial Nets (GANs)” (https://arxiv.org/abs/1703.00573)

    Notes
    -----
    This class is a subclass of TorchModel.  It accepts all the keyword arguments
    from TorchModel.
    """

    def __init__(self,
                 n_generators: int = 1,
                 n_discriminators: int = 1,
                 create_discriminator_loss: Optional[Callable] = None,
                 create_generator_loss: Optional[Callable] = None,
                 _call_discriminator: Optional[Callable] = None,
                 device: Optional[torch.device] = None,
                 **kwargs):
        """
        Parameters
        ----------
        n_generators: int
            the number of generators to include
        n_discriminators: int
            the number of discriminators to include
        create_discriminator_loss: Callable
            a function that returns the loss function for the discriminator.  It will
            be called with two arguments: the output from the discriminator on a
            batch of training data, and the output from the discriminator on a batch
            of generated data.  The default implementation is appropriate for most
            cases.  Subclasses can override this if the need to customize it.
        create_generator_loss: Callable
            a function that returns the loss function for the generator.  It will be
            called with one argument: the output from the discriminator on a batch of
            generated data.  The default implementation is appropriate for most
            cases.  Subclasses can override this if the need to customize it.
        _call_discriminator: Callable
            a function that invokes the discriminator on a set of inputs.  It will be
            called with three arguments: the discriminator to invoke, the list of
            data inputs, and the list of conditional inputs.  The default
            implementation is appropriate for most cases.  Subclasses can override
            this if the need to customize it.
        """
        self.n_generators = n_generators
        self.n_discriminators = n_discriminators

        if device is None:
            if torch.cuda.is_available():
                self.device = torch.device('cuda')
            elif torch.backends.mps.is_available():
                self.device = torch.device('mps')
            else:
                self.device = torch.device('cpu')
        model = GAN(noise_input_shape=self.get_noise_input_shape(),
                    data_input_shape=self.get_data_input_shapes(),
                    conditional_input_shape=self.get_conditional_input_shapes(),
                    generator_fn=self.create_generator,
                    discriminator_fn=self.create_discriminator,
                    device=self.device,
                    n_generators=n_generators,
                    n_discriminators=n_discriminators,
                    create_discriminator_loss=create_discriminator_loss,
                    create_generator_loss=create_generator_loss,
                    _call_discriminator=_call_discriminator)
        self.discrim_loss_fn = model.discrim_loss_fn
        self.gen_loss_fn = model.gen_loss_fn
        self.data_inputs = model.data_inputs
        self.conditional_inputs = model.conditional_inputs
        self.generators = model.generators
        self.discriminators = model.discriminators
        self.gen_variables = model.gen_variables
        self.discrim_variables = model.discrim_variables
        self.get_noise_batch = model.get_noise_batch

        super(GANModel, self).__init__(model, loss=self.gen_loss_fn, **kwargs)

    def get_noise_input_shape(self):
        """Get the shape of the generator's noise input layer.

        Subclasses must override this to return a tuple giving the shape of the
        noise input.  The actual Input layer will be created automatically.  The
        dimension corresponding to the batch size should be omitted.
        """
        raise NotImplementedError("Subclasses must implement this.")

    def get_data_input_shapes(self):
        """Get the shapes of the inputs for training data.

        Subclasses must override this to return a list of tuples, each giving the
        shape of one of the inputs.  The actual Input layers will be created
        automatically.  This list of shapes must also match the shapes of the
        generator's outputs.  The dimension corresponding to the batch size should
        be omitted.
        """
        raise NotImplementedError("Subclasses must implement this.")

    def get_conditional_input_shapes(self):
        """Get the shapes of any conditional inputs.

        Subclasses may override this to return a list of tuples, each giving the
        shape of one of the conditional inputs.  The actual Input layers will be
        created automatically.  The dimension corresponding to the batch size should
        be omitted.

        The default implementation returns an empty list, meaning there are no
        conditional inputs.
        """
        return []

    def create_generator(self):
        """Create and return a generator.

        Subclasses must override this to construct the generator.  The returned
        value should be a tf.keras.Model whose inputs are a batch of noise, followed
        by any conditional inputs.  The number and shapes of its outputs must match
        the return value from get_data_input_shapes(), since generated data must
        have the same form as training data.
        """
        raise NotImplementedError("Subclasses must implement this.")

    def create_discriminator(self):
        """Create and return a discriminator.

        Subclasses must override this to construct the discriminator.  The returned
        value should be a tf.keras.Model whose inputs are all data inputs, followed
        by any conditional inputs.  Its output should be a one dimensional tensor
        containing the probability of each sample being a training sample.
        """
        raise NotImplementedError("Subclasses must implement this.")

    def fit_gan(self,
                batches,
                generator_steps=1,
                max_checkpoints_to_keep=5,
                checkpoint_interval=1000,
                restore=False) -> None:
        """Train this model on data.

        Parameters
        ----------
        batches: iterable
            batches of data to train the discriminator on, each represented as a dict
            that maps Inputs to values.  It should specify values for all members of
            data_inputs and conditional_inputs.
        generator_steps: float
            the number of training steps to perform for the generator for each batch.
            This can be used to adjust the ratio of training steps for the generator
            and discriminator.  For example, 2.0 will perform two training steps for
            every batch, while 0.5 will only perform one training step for every two
            batches.
        max_checkpoints_to_keep: int
            the maximum number of checkpoints to keep.  Older checkpoints are discarded.
        checkpoint_interval: int
            the frequency at which to write checkpoints, measured in batches.  Set
            this to 0 to disable automatic checkpointing.
        restore: bool
            if True, restore the model from the most recent checkpoint before training
            it.
        """
        self._ensure_built()

        gen_train_fraction = 0.0
        discrim_error = 0.0
        gen_error = 0.0
        discrim_average_steps = 0
        gen_average_steps = 0
        time1 = time.time()

        for feed_dict in batches:
            # Every call to fit_generator() will increment global_step, but we only
            # want it to get incremented once for the entire batch, so record the
            # value and keep resetting it.

            global_step = self.get_global_step()

            # Train the discriminator.

            inputs = [self.get_noise_batch(self.batch_size)]

            for input in self.data_inputs:
                inputs.append(feed_dict[input])
            for input in self.conditional_inputs:
                inputs.append(feed_dict[input])
            discrim_error += self.fit_generator(
                [(inputs, [], [])],
                variables=self.discrim_variables,
                loss=self.discrim_loss_fn,
                checkpoint_interval=0,
                restore=restore)
            restore = False
            discrim_average_steps += 1

            # Train the generator.

            if generator_steps > 0.0:
                gen_train_fraction += generator_steps
                while gen_train_fraction >= 1.0:
                    inputs = [self.get_noise_batch(self.batch_size)
                             ] + inputs[1:]

                    gen_error += self.fit_generator(
                        [(inputs, [], [])],
                        variables=self.gen_variables,
                        loss=self.gen_loss_fn,
                        checkpoint_interval=0)
                    gen_average_steps += 1
                    gen_train_fraction -= 1.0
            self._global_step = global_step + 1

            # Write checkpoints and report progress.

            if discrim_average_steps == checkpoint_interval:
                self.save_checkpoint(max_checkpoints_to_keep)
                discrim_loss = discrim_error / max(1, discrim_average_steps)
                gen_loss = gen_error / max(1, gen_average_steps)
                print(
                    'Ending global_step %d: generator average loss %g, discriminator average loss %g'
                    % (global_step, gen_loss, discrim_loss))
                discrim_error = 0.0
                gen_error = 0.0
                discrim_average_steps = 0
                gen_average_steps = 0

        # Write out final results.

        if checkpoint_interval > 0:
            if discrim_average_steps > 0 and gen_average_steps > 0:
                discrim_loss = discrim_error / discrim_average_steps
                gen_loss = gen_error / gen_average_steps
                print(
                    'Ending global_step %d: generator average loss %g, discriminator average loss %g'
                    % (global_step, gen_loss, discrim_loss))
            self.save_checkpoint(max_checkpoints_to_keep)
            time2 = time.time()
            print("TIMING: model fitting took %0.3f s" % (time2 - time1))

    def predict_gan_generator(self,
                              batch_size=1,
                              noise_input=None,
                              conditional_inputs=[],
                              generator_index=0):
        """Use the GAN to generate a batch of samples.

        Parameters
        ----------
        batch_size: int
            the number of samples to generate.  If either noise_input or
            conditional_inputs is specified, this argument is ignored since the batch
            size is then determined by the size of that argument.
        noise_input: array
            the value to use for the generator's noise input.  If None (the default),
            get_noise_batch() is called to generate a random input, so each call will
            produce a new set of samples.
        conditional_inputs: list of arrays
            the values to use for all conditional inputs.  This must be specified if
            the GAN has any conditional inputs.
        generator_index: int
            the index of the generator (between 0 and n_generators-1) to use for
            generating the samples.

        Returns
        -------
        An array (if the generator has only one output) or list of arrays (if it has
        multiple outputs) containing the generated samples.
        """
        if noise_input is not None:
            batch_size = len(noise_input)
        elif len(conditional_inputs) > 0:
            batch_size = len(conditional_inputs[0])
        if noise_input is None:
            noise_input = self.get_noise_batch(batch_size)
        inputs = [noise_input]
        inputs += conditional_inputs
        inputs = [i.astype(np.float32) for i in inputs]
        inputs = [torch.from_numpy(i).to(self.device) for i in inputs]
        pred = self.generators[generator_index](_list_or_tensor(inputs))
        pred = pred.cpu().detach().numpy()
        return pred


class WGANModel(GANModel):
    """Implements Wasserstein Generative Adversarial Networks.

    This class implements Wasserstein Generative Adversarial Networks (WGANs) as
    described in Arjovsky et al., "Wasserstein GAN" [1]_.
    A WGAN is conceptually rather different from a conventional GAN, but in
    practical terms very similar.  It reinterprets the discriminator (often called
    the "critic" in this context) as learning an approximation to the Earth Mover
    distance between the training and generated distributions.  The generator is
    then trained to minimize that distance.  In practice, this just means using
    slightly different loss functions for training the generator and discriminator.

    WGANs have theoretical advantages over conventional GANs, and they often work
    better in practice.  In addition, the discriminator's loss function can be
    directly interpreted as a measure of the quality of the model.  That is an
    advantage over conventional GANs, where the loss does not directly convey
    information about the quality of the model.

    The theory WGANs are based on requires the discriminator's gradient to be
    bounded.  The original paper achieved this by clipping its weights.  This
    class instead does it by adding a penalty term to the discriminator's loss, as
    described in [2]_.  This is sometimes found to produce better results.

    There are a few other practical differences between GANs and WGANs.  In a
    conventional GAN, the discriminator's output must be between 0 and 1 so it can
    be interpreted as a probability.  In a WGAN, it should produce an unbounded
    output that can be interpreted as a distance.

    When training a WGAN, you also should usually use a smaller value for
    generator_steps.  Conventional GANs rely on keeping the generator and
    discriminator "in balance" with each other.  If the discriminator ever gets
    too good, it becomes impossible for the generator to fool it and training
    stalls.  WGANs do not have this problem, and in fact the better the
    discriminator is, the easier it is for the generator to improve.  It therefore
    usually works best to perform several training steps on the discriminator for
    each training step on the generator.

    Examples
    --------
    Importing necessary modules

    >>> import deepchem as dc
    >>> from deepchem.models.torch_models.gan import WGANModel
    >>> import torch
    >>> import torch.nn as nn
    >>> import torch.nn.functional as F

    Creating a Generator

    >>> class Generator(nn.Module):
    ...     def __init__(self, noise_input_shape, conditional_input_shape):
    ...         super(Generator, self).__init__()
    ...         self.noise_input_shape = noise_input_shape
    ...         self.conditional_input_shape = conditional_input_shape
    ...         self.noise_dim = noise_input_shape[1:]
    ...         self.conditional_dim = conditional_input_shape[1:]
    ...         input_dim = sum(self.noise_dim) + sum(self.conditional_dim)
    ...         self.output = nn.Linear(input_dim, 1)
    ...     def forward(self, input):
    ...         noise_input, conditional_input = input
    ...         inputs = torch.cat((noise_input, conditional_input), dim=1)
    ...         output = self.output(inputs)
    ...         return output

    Creating a Discriminator

    >>> class Discriminator(nn.Module):
    ...     def __init__(self, data_input_shape, conditional_input_shape):
    ...         super(Discriminator, self).__init__()
    ...         self.data_input_shape = data_input_shape
    ...         self.conditional_input_shape = conditional_input_shape
    ...         # Extracting the actual data dimension
    ...         data_dim = data_input_shape[1:]
    ...         # Extracting the actual conditional dimension
    ...         conditional_dim = conditional_input_shape[1:]
    ...         input_dim = sum(data_dim) + sum(conditional_dim)
    ...         # Define the dense layers
    ...         self.dense1 = nn.Linear(input_dim, 10)
    ...         self.dense2 = nn.Linear(10, 1)
    ...     def forward(self, input):
    ...         data_input, conditional_input = input
    ...         # Concatenate data_input and conditional_input along the second dimension
    ...         discrim_in = torch.cat((data_input, conditional_input), dim=1)
    ...         # Pass the concatenated input through the dense layers
    ...         x = F.relu(self.dense1(discrim_in))
    ...         output = self.dense2(x)
    ...         return output

    Creating an Example WGANModel class

    >>> class ExampleWGAN(WGANModel):
    ...     def get_noise_input_shape(self):
    ...         return (100,2,)
    ...     def get_data_input_shapes(self):
    ...         return [(100,1,)]
    ...     def get_conditional_input_shapes(self):
    ...         return [(100,1,)]
    ...     def create_generator(self):
    ...         noise_dim = self.get_noise_input_shape()
    ...         conditional_dim = self.get_conditional_input_shapes()[0]
    ...         return nn.Sequential(Generator(noise_dim, conditional_dim))
    ...     def create_discriminator(self):
    ...         data_input_shape = self.get_data_input_shapes()[0]
    ...         conditional_input_shape = self.get_conditional_input_shapes()[0]
    ...         return nn.Sequential(
    ...             Discriminator(data_input_shape, conditional_input_shape))

    Defining a function to generate data

    >>> def generate_batch(batch_size):
    ...     means = 10 * np.random.random([batch_size, 1])
    ...     values = np.random.normal(means, scale=2.0)
    ...     return means, values
    >>> def generate_data(gan, batches, batch_size):
    ...     for _ in range(batches):
    ...         means, values = generate_batch(batch_size)
    ...         batch = {
    ...             gan.data_inputs[0]: values,
    ...             gan.conditional_inputs[0]: means
    ...         }
    ...         yield batch

    Defining the WGANModel

    >>> wgan = ExampleWGAN(learning_rate=0.01,
    ...               gradient_penalty=0.1)
    >>> data = generate_data(wgan, 500, 100)
    >>> wgan.fit_gan(data, generator_steps=0.1, checkpoint_interval=0)
    >>> means = 10 * np.random.random([1000, 1])
    >>> values = wgan.predict_gan_generator(conditional_inputs=[means])

    References
    ----------
    .. [1] Arjovsky, Martin, Soumith Chintala, and Léon Bottou.
        "Wasserstein generative adversarial networks."
        International conference on machine learning. PMLR, 2017.
        (https://arxiv.org/abs/1701.07875)

    .. [2] Gulrajani, Ishaan, et al. "Improved training of wasserstein gans."
        Advances in neural information processing systems 30 (2017).
        (https://arxiv.org/abs/1704.00028)
    """

    def __init__(self, gradient_penalty: float = 10.0, **kwargs):
        """Construct a WGAN.

        In addition to the following, this class accepts all the keyword arguments
        from GAN and TorchModel.

        Parameters
        ----------
        gradient_penalty: float default 10.0
            the magnitude of the gradient penalty loss
        """
        self.gradient_penalty = gradient_penalty
        super(WGANModel, self).__init__(
            create_discriminator_loss=self.create_discriminator_loss,
            create_generator_loss=self.create_generator_loss,
            _call_discriminator=self._call_discriminator,
            **kwargs)

    def _call_discriminator(self, discriminator: nn.Module, inputs: List,
                            conditional_input_layers: List, train: bool):
        """ Invoke the discriminator on a set of inputs.

        Parameters
        ----------
        discriminator: nn.Module
            the discriminator to invoke
        inputs: list of Tensor
            the inputs to the discriminator.  The first element must be a batch of
            data, followed by any conditional inputs.
        conditional_input_layers: list of Tensor
            the conditional inputs to the discriminator.  These are needed
            because the discriminator may have multiple outputs, and the conditional
            inputs must be passed to all of them.
        train: bool
            if True, the discriminator should be invoked in training mode. If False,
            it should be invoked in inference mode.

        Returns
        -------
        the output from the discriminator
        """
        if train:
            penalty = GradientPenaltyLayer(self, discriminator)
            return penalty(inputs, conditional_input_layers)
        return discriminator(_list_or_tensor(inputs + conditional_input_layers))

    def create_generator_loss(self,
                              discrim_output: torch.Tensor) -> torch.Tensor:
        """Create the loss function for the generator.

        Parameters
        ----------
        discrim_output : torch.Tensor
            the output from the discriminator on a batch of generated data.  This is
            its estimate of the probability that each sample is training data.

        Returns
        -------
        torch.Tensor
            A Tensor equal to the mean of the inputs
        """
        return torch.mean(discrim_output)

    def create_discriminator_loss(
            self, discrim_output_train: List[torch.Tensor],
            discrim_output_gen: torch.Tensor) -> torch.Tensor:
        """Create the loss function for the discriminator.

        Parameters
        ----------
        discrim_output_train : List[Tensor]
            the output from the discriminator on a batch of training data.  This is
            its estimate of the probability that each sample is training data.
        discrim_output_gen : Tensor
            the output from the discriminator on a batch of generated data.

        Returns
        -------
        torch.Tensor
            A Tensor equal to the loss function to use for optimizing the discriminator.
        """
        return torch.mean(discrim_output_train[0][0] -
                          discrim_output_gen[1]) + discrim_output_train[1]


class GradientPenaltyLayer(nn.Module):
    """Implements the gradient penalty loss term for WGANs.

    This class implements the gradient penalty loss term for WGANs as described in
    Gulrajani et al., "Improved Training of Wasserstein GANs" [1]_.  It is used
    internally by WGANModel

    Examples
    --------
    Importing necessary modules

    >>> import deepchem
    >>> from deepchem.models.torch_models.gan import WGANModel
    >>> from deepchem.models.torch_models import GradientPenaltyLayer
    >>> import torch
    >>> import torch.nn as nn
    >>> import torch.nn.functional as F

    Creating a Generator

    >>> class Generator(nn.Module):
    ...     def __init__(self, noise_input_shape, conditional_input_shape):
    ...         super(Generator, self).__init__()
    ...         self.noise_input_shape = noise_input_shape
    ...         self.conditional_input_shape = conditional_input_shape
    ...         self.noise_dim = noise_input_shape[1:]
    ...         self.conditional_dim = conditional_input_shape[1:]
    ...         input_dim = sum(self.noise_dim) + sum(self.conditional_dim)
    ...         self.output = nn.Linear(input_dim, 1)
    ...     def forward(self, input):
    ...         noise_input, conditional_input = input
    ...         inputs = torch.cat((noise_input, conditional_input), dim=1)
    ...         output = self.output(inputs)
    ...         return output

    Creating a Discriminator

    >>> class Discriminator(nn.Module):
    ...     def __init__(self, data_input_shape, conditional_input_shape):
    ...         super(Discriminator, self).__init__()
    ...         self.data_input_shape = data_input_shape
    ...         self.conditional_input_shape = conditional_input_shape
    ...         # Extracting the actual data dimension
    ...         data_dim = data_input_shape[1:]
    ...         # Extracting the actual conditional dimension
    ...         conditional_dim = conditional_input_shape[1:]
    ...         input_dim = sum(data_dim) + sum(conditional_dim)
    ...         # Define the dense layers
    ...         self.dense1 = nn.Linear(input_dim, 10)
    ...         self.dense2 = nn.Linear(10, 1)
    ...     def forward(self, input):
    ...         data_input, conditional_input = input
    ...         # Concatenate data_input and conditional_input along the second dimension
    ...         discrim_in = torch.cat((data_input, conditional_input), dim=1)
    ...         # Pass the concatenated input through the dense layers
    ...         x = F.relu(self.dense1(discrim_in))
    ...         output = self.dense2(x)
    ...         return output

    Creating an Example WGANModel class

    >>> class ExampleWGAN(WGANModel):
    ...     def get_noise_input_shape(self):
    ...         return (100,2,)
    ...     def get_data_input_shapes(self):
    ...         return [(100,1,)]
    ...     def get_conditional_input_shapes(self):
    ...         return [(100,1,)]
    ...     def create_generator(self):
    ...         noise_dim = self.get_noise_input_shape()
    ...         conditional_dim = self.get_conditional_input_shapes()[0]
    ...         return nn.Sequential(Generator(noise_dim, conditional_dim))
    ...     def create_discriminator(self):
    ...         data_input_shape = self.get_data_input_shapes()[0]
    ...         conditional_input_shape = self.get_conditional_input_shapes()[0]
    ...         return nn.Sequential(
    ...             Discriminator(data_input_shape, conditional_input_shape))

    Defining an Example GradientPenaltyLayer

    >>> wgan = ExampleWGAN()
    >>> discriminator = wgan.discriminators[0]
    >>> gpl = GradientPenaltyLayer(wgan, discriminator)
    >>> inputs = [torch.randn(4, 1)]
    >>> conditional_inputs = [torch.randn(4, 1)]
    >>> output, penalty = gpl(inputs, conditional_inputs)


    References
    ----------
    .. [1]_ Gulrajani, Ishaan, et al. "Improved training of wasserstein gans."
        Advances in neural information processing systems 30 (2017).
        (https://arxiv.org/abs/1704.00028)
    """

    def __init__(self, gan: WGANModel, discriminator: nn.Module,
                 **kwargs) -> None:
        """Construct a GradientPenaltyLayer.

        Parameters
        ----------
        gan : WGANModel
            the WGANModel that this layer is part of
        discriminator : nn.Module
            the discriminator to compute the gradient penalty for
        """
        super(GradientPenaltyLayer, self).__init__(**kwargs)
        self.gan = gan
        self.discriminator = discriminator

    def forward(self, inputs: Union[list, torch.Tensor],
                conditional_inputs: torch.Tensor) -> list:
        """ Compute the output of the gradient penalty layer.

        Parameters
        ----------
        inputs: list of Tensor
            the inputs to the discriminator.
        conditional_inputs: Tensor
            the conditional inputs to the discriminator.

        Returns
        -------
        output: list [Tensor, Tensor]
            the output from the discriminator, followed by the gradient penalty.
        """
        input_on_device = []
        for tensor in inputs:
            tensor = tensor.to(torch.float32).to(self.gan.device)
            input_on_device.append(
                tensor.requires_grad_(True).to(self.gan.device))
        conditional_inputs_on_device = []
        for tensor in conditional_inputs:
            tensor = tensor.to(torch.float32).to(self.gan.device)
            conditional_inputs_on_device.append(tensor.to(self.gan.device))
        output = self.discriminator(
            _list_or_tensor(input_on_device + conditional_inputs_on_device)).to(
                self.gan.device)
        gradients_raw = torch.autograd.grad(
            outputs=output,
            inputs=input_on_device,
            grad_outputs=torch.ones_like(output),
            create_graph=True,
            allow_unused=True)
        gradients = [g for g in gradients_raw if g is not None]
        penalty: Union[torch.Tensor, float]
        norm2: Union[float, torch.Tensor]
        if gradients:
            norm2 = 0.0
            for g in gradients:
                g2 = torch.square(g)
                dims = len(list(g.shape))
                if dims > 1:
                    g2 = torch.sum(g2, dim=list(range(1, dims)))
                norm2 += g2  # type: ignore
            penalty = torch.square(torch.sqrt(norm2) - 1.0)  # type: ignore
            penalty = self.gan.gradient_penalty * torch.mean(penalty)
        else:
            penalty = 0.0
        return [output, penalty]



import numpy as np
from typing import DefaultDict, Optional

from deepchem.utils.typing import PymatgenComposition
from deepchem.feat import MaterialCompositionFeaturizer

elements_tl = [
    'H', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Na', 'Mg', 'Al', 'Si', 'P', 'S',
    'Cl', 'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn',
    'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr', 'Nb', 'Mo', 'Tc',
    'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te', 'I', 'Xe', 'Cs', 'Ba',
    'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er',
    'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg', 'Tl',
    'Pb', 'Bi', 'Ac', 'Th', 'Pa', 'U', 'Np', 'Pu'
]


class ElemNetFeaturizer(MaterialCompositionFeaturizer):
    """
    Fixed size vector of length 86 containing raw fractional elemental
    compositions in the compound. The 86 chosen elements are based on the
    original implementation at https://github.com/NU-CUCIS/ElemNet.

    Returns a vector containing fractional compositions of each element
    in the compound.

    References
    ----------
    .. [1] Jha, D., Ward, L., Paul, A. et al. Sci Rep 8, 17593 (2018).
        https://doi.org/10.1038/s41598-018-35934-y

    Examples
    --------
    >>> import deepchem as dc
    >>> comp = "Fe2O3"
    >>> featurizer = dc.feat.ElemNetFeaturizer()
    >>> features = featurizer.featurize([comp])
    >>> type(features[0])
    <class 'numpy.ndarray'>
    >>> features[0].shape
    (86,)
    >>> round(sum(features[0]))
    1

    Note
    ----
    This class requires Pymatgen to be installed.
    """

    def get_vector(self, comp: DefaultDict) -> Optional[np.ndarray]:
        """
        Converts a dictionary containing element names and corresponding
        compositional fractions into a vector of fractions.

        Parameters
        ----------
        comp: collections.defaultdict object
            Dictionary mapping element names to fractional compositions.

        Returns
        -------
        fractions: np.ndarray
            Vector of fractional compositions of each element.
        """
        if all(e in elements_tl for e in comp):
            return np.array([comp[e] if e in comp else 0 for e in elements_tl],
                            np.float32)
        else:
            return None

    def _featurize(self, datapoint: PymatgenComposition,
                   **kwargs) -> Optional[np.ndarray]:
        """
        Calculate 86 dimensional vector containing fractional compositions of
        each element in the compound.

        Parameters
        ----------
        datapoint: pymatgen.core.Composition object
            Composition object.

        Returns
        -------
        feats: np.ndarray
            86 dimensional vector containing fractional compositions of elements.
        """
        if 'composition' in kwargs and datapoint is None:
            datapoint = kwargs.get("composition")
            raise DeprecationWarning(
                'Composition is being phased out as a parameter, please pass "datapoint" instead.'
            )

        fractions = datapoint.fractional_composition.get_el_amt_dict()
        return self.get_vector(fractions)



import math
import torch
from typing import Union, Optional, Tuple

eps = 1e-12
ZType = Union[int, float, torch.Tensor]

# safe operations


def safepow(a: torch.Tensor,
            p: torch.Tensor,
            eps: float = 1e-12) -> torch.Tensor:
    """Safely calculate the power of a tensor with a small eps to avoid nan.

    Examples
    --------
    >>> import torch
    >>> a = torch.tensor([1e-35, 2e-40])
    >>> p = torch.tensor([2., 3])
    >>> safepow(a, p)
    tensor([1.0000e-24, 1.0000e-36])
    >>> a**p
    tensor([0., 0.])

    Parameters
    ----------
    a: torch.Tensor
        Base tensor on which to calculate the power. Must be positive.
    p: torch.Tensor
        Power tensor, by which to calculate the power.
    eps: float (default 1e-12)
        The eps to add to the base tensor.

    Returns
    -------
    torch.Tensor
        The result tensor.

    Raises
    ------
    RuntimeError
        If the base tensor contains negative values.

    """
    if torch.any(a < 0):
        raise RuntimeError("safepow only works for positive base")
    base = torch.sqrt(a * a + eps * eps)  # soft clip
    return base**p


def safenorm(a: torch.Tensor, dim: int, eps: float = 1e-15) -> torch.Tensor:
    """
    Calculate the 2-norm safely. The square root of the inner product of a
    vector with itself.

    Examples
    --------
    >>> import torch
    >>> a = torch.tensor([1e-35, 2e-40])
    >>> safenorm(a, 0)
    tensor(1.4142e-15)
    >>> a.norm()
    tensor(0.)


    Parameters
    ----------
    a: torch.Tensor
        The tensor to calculate the norm.
    dim: int
        The dimension to calculate the norm.
    eps: float (default 1e-15)
        The eps to add to the base tensor.

    Returns
    -------
    torch.Tensor
        The result tensor.

    """
    return torch.sqrt(torch.sum(a * a + eps * eps, dim=dim))


# occupation number gradients


def occnumber(
    a: ZType,
    n: Optional[int] = None,
    dtype: torch.dtype = torch.double,
    device: torch.device = torch.device('cpu')
) -> torch.Tensor:
    """Occupation number (maxed at 1) where the total sum of the
    output equals to a with length of the output is n.

    Examples
    --------
    >>> import torch
    >>> occnumber(torch.tensor(2.5), 3, torch.double, torch.device('cpu'))
    tensor([1.0000, 1.0000, 0.5000], dtype=torch.float64)
    >>> occnumber(2.5)
    tensor([1.0000, 1.0000, 0.5000], dtype=torch.float64)

    Parameters
    ----------
    a: ZType
        Total sum of the output
    n: Optional[int] (default None)
        Length of the output
    dtype: torch.dtype (default torch.double)
        Data type of the output
    device: torch.device (default torch.device('cpu'))
        Device of the output

    Returns
    -------
    torch.Tensor
        The constructed occupation number

    """

    if isinstance(a, torch.Tensor):
        assert a.numel() == 1
        floor_a, ceil_a = get_floor_and_ceil(a.item())
    else:  # int or float
        floor_a, ceil_a = get_floor_and_ceil(a)

    # get the length of the tensor output
    if n is None:
        nlength = ceil_a
    else:
        nlength = n
        assert nlength >= ceil_a, "The length of occupation number must be at least %d" % ceil_a

    if isinstance(a, torch.Tensor):
        res = _OccNumber.apply(a, floor_a, ceil_a, nlength, dtype, device)
    else:
        res = _construct_occ_number(a,
                                    floor_a,
                                    ceil_a,
                                    nlength,
                                    dtype=dtype,
                                    device=device)
    return res


def _construct_occ_number(a: float, floor_a: int, ceil_a: int, nlength: int,
                          dtype: torch.dtype,
                          device: torch.device) -> torch.Tensor:
    """Construct the occupation number (maxed at 1) where the total sum of the
    output equals to a with length of the output is nlength.

    Examples
    --------
    >>> import torch
    >>> _construct_occ_number(2.5, 2, 3, 3, torch.double, torch.device('cpu'))
    tensor([1.0000, 1.0000, 0.5000], dtype=torch.float64)

    Parameters
    ----------
    a: float
        Total sum of the output
    floor_a: int
        Floor of a
    ceil_a: int
        Ceiling of a
    nlength: int
        Length of the output
    dtype: torch.dtype
        Data type of the output
    device: torch.device
        Device of the output

    Returns
    -------
    torch.Tensor
        The constructed occupation number

    """
    res = torch.zeros(nlength, dtype=dtype, device=device)
    res[:floor_a] = 1
    if ceil_a > floor_a:
        res[ceil_a - 1] = a - floor_a
    return res


class _OccNumber(torch.autograd.Function):
    """Construct the occupation number (maxed at 1) where the total sum of the
    output equals to a with length of the output is nlength.

    Examples
    --------
    >>> import torch
    >>> a = torch.tensor(2.5)
    >>> _OccNumber.apply(a, 2, 3, 3, torch.double, torch.device('cpu'))
    tensor([1.0000, 1.0000, 0.5000], dtype=torch.float64)

    """

    @staticmethod
    def forward(  # type: ignore
            self, a: torch.Tensor, floor_a: int, ceil_a: int, nlength: int,
            dtype: torch.dtype, device: torch.device) -> torch.Tensor:
        """Forward pass of the occupation number.

        Parameters
        ----------
        a: torch.Tensor
            Total sum of the output
        floor_a: int
            Floor of a
        ceil_a: int
            Ceiling of a
        nlenght: int
            Length of the output
        dtype: torch.dtype
            Data type of the output
        device: torch.device
            Device of the output

        Returns
        -------
        torch.Tensor
            The constructed occupation number

        """
        res = _construct_occ_number(float(a.item()),
                                    floor_a,
                                    ceil_a,
                                    nlength,
                                    dtype=dtype,
                                    device=device)
        self.ceil_a = ceil_a
        return res

    @staticmethod
    def backward(self, grad_res: torch.Tensor):  # type: ignore
        """Backward pass of the occupation number.

        Parameters
        ----------
        grad_res: torch.Tensor
            Gradient of the output

        Returns
        -------
        Tuple[torch.Tensor, None, None, None, None, None]
            Gradient of the input

        """
        grad_a = grad_res[self.ceil_a - 1]
        return (grad_a,) + (None,) * 5


def get_floor_and_ceil(aa: Union[int, float]) -> Tuple[int, int]:
    """get the ceiling and flooring of aa.

    Examples
    --------
    >>> get_floor_and_ceil(2.5)
    (2, 3)

    Parameters
    ----------
    aa: Union[int, float]
        The input number

    Returns
    -------
    Tuple[int, int]
        The flooring and ceiling of aa

    """
    if isinstance(aa, int):
        ceil_a: int = aa
        floor_a: int = aa
    else:  # floor
        ceil_a = int(math.ceil(aa))
        floor_a = int(math.floor(aa))
    return floor_a, ceil_a


# other tensor ops


def safe_cdist(a: torch.Tensor,
               b: torch.Tensor,
               add_diag_eps: bool = False,
               diag_inf: bool = False):
    """L2 pairwise distance of a and b. The diagonal is either
    replaced with a small eps or infinite.

    Examples
    --------
    >>> import torch
    >>> a = torch.tensor([[1., 2], [3, 4]])
    >>> b = torch.tensor([[1., 2], [3, 4]])
    >>> safe_cdist(a, b)
    tensor([[0.0000, 2.8284],
            [2.8284, 0.0000]])
    >>> safe_cdist(a, b, add_diag_eps=True)
    tensor([[1.4142e-12, 2.8284e+00],
            [2.8284e+00, 1.4142e-12]])
    >>> safe_cdist(a, b, diag_inf=True)
    tensor([[   inf, 2.8284],
            [2.8284,    inf]])

    Parameters
    ----------
    a: torch.Tensor
        First Tensor. Shape: (*BA, na, ndim)
    n: torch.Tensor
        Second Tensor. Shape: (*BB, nb, ndim)

    Returns
    -------
    torch.Tensor
        Pairwise distance. Shape: (*BAB, na, nb)

    """
    square_mat = a.shape[-2] == b.shape[-2]

    dtype = a.dtype
    device = a.device
    ab = a.unsqueeze(-2) - b.unsqueeze(-3)  # (*BAB, na, nb, ndim)

    # add the diagonal with a small eps to safeguard from nan
    if add_diag_eps:
        if not square_mat:
            raise ValueError(
                "Enabling add_diag_eps for non-square result matrix is invalid")
        ab = ab + torch.eye(ab.shape[-2], dtype=dtype,
                            device=device).unsqueeze(-1) * eps

    ab = ab.norm(dim=-1)  # (*BAB, na, nb)

    # replace the diagonal with infinite (usually used for coulomb matrix)
    if diag_inf:
        if not square_mat:
            raise ValueError(
                "Enabling diag_inf for non-square result matrix is invalid")

        infdiag = torch.eye(ab.shape[-1], dtype=dtype, device=device)
        idiag = infdiag.diagonal()
        idiag[:] = float("inf")
        ab = ab + infdiag

    return ab


def safedenom(r: torch.Tensor, eps: float) -> torch.Tensor:
    """Avoid division by zero by replacing zero elements with eps.

    Used in CG and BiCGStab.

    Examples
    --------
    >>> import torch
    >>> r = torch.tensor([1e-11, 0])
    >>> safedenom(r, 1e-12)
    tensor([1.0000e-11, 1.0000e-12])

    Parameters
    ----------
    r: torch.Tensor
        The residual vector
    eps: float
        The minimum value to avoid division by zero

    Returns
    -------
    r: torch.Tensor
        The residual vector with zero elements replaced by eps

    """
    r[r == 0] = eps
    return r



import logging
import time
from collections.abc import Sequence as SequenceCollection
from typing import (TYPE_CHECKING, Any, Callable, Iterable, List, Optional,
                    Tuple, Union)

import numpy as np
import torch
from deepchem.models.optimizers import LearningRateSchedule
from deepchem.models.torch_models import TorchModel
from deepchem.trans import Transformer, undo_transforms
from deepchem.utils.typing import LossFn, OneOrMany
from transformers.data.data_collator import DataCollatorForLanguageModeling
from transformers.models.auto import AutoModel, AutoModelForSequenceClassification, AutoModelForMaskedLM

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    import transformers
    from transformers.modeling_utils import PreTrainedModel


class HuggingFaceModel(TorchModel):
    r"""Wrapper class that wraps HuggingFace models as DeepChem models

    The class provides a wrapper for wrapping models from HuggingFace
    ecosystem in DeepChem and training it via DeepChem's api. The reason
    for this might be that you might want to do an apples-to-apples comparison
    between HuggingFace from the transformers library and DeepChem library.

    The `HuggingFaceModel` has a Has-A relationship by wrapping models from
    `transformers` library. Once a model is wrapped, DeepChem's API are used
    for training, prediction, evaluation and other downstream tasks.

    A `HuggingFaceModel` wrapper also has a `tokenizer` which tokenizes raw
    SMILES strings into tokens to be used by downstream models.  The SMILES
    strings are generally stored in the `X` attribute of deepchem.data.Dataset object'.
    This differs from the DeepChem standard workflow as tokenization is done
    on the fly here. The approach allows us to leverage `transformers` library's fast
    tokenization algorithms and other utilities like data collation, random masking of tokens
    for masked language model training etc.


    Parameters
    ----------
    model: transformers.modeling_utils.PreTrainedModel
        The HuggingFace model to wrap.
    task: str, (optional, default None)
        The task defines the type of learning task in the model. The supported tasks are
         - `mlm` - masked language modeling commonly used in pretraining
         - `mtr` - multitask regression - a task used for both pretraining base models and finetuning
         - `regression` - use it for regression tasks, like property prediction
         - `classification` - use it for classification tasks
        When the task is not specified or None, the wrapper returns raw output of the HuggingFaceModel.
        In cases where the HuggingFaceModel is a model without a task specific head, this output will be
        the last hidden states.
    tokenizer: transformers.tokenization_utils.PreTrainedTokenizer
        Tokenizer

    Example
    -------
    >>> import os
    >>> import tempfile
    >>> tempdir = tempfile.mkdtemp()

    >>> # preparing dataset
    >>> smiles = ['CN(c1ccccc1)c1ccccc1C(=O)NCC1(O)CCOCC1', 'CC[NH+](CC)C1CCC([NH2+]C2CC2)(C(=O)[O-])C1', \
    ...     'COCC(CNC(=O)c1ccc2c(c1)NC(=O)C2)OC', 'OCCn1cc(CNc2cccc3c2CCCC3)nn1', \
    ...     'CCCCCCc1ccc(C#Cc2ccc(C#CC3=CC=C(CCC)CC3)c(C3CCCCC3)c2)c(F)c1', 'nO=C(NCc1ccc(F)cc1)N1CC=C(c2c[nH]c3ccccc23)CC1']
    >>> filepath = os.path.join(tempdir, 'smiles.txt')
    >>> f = open(filepath, 'w')
    >>> f.write('\n'.join(smiles))
    253
    >>> f.close()

    >>> # preparing tokenizer
    >>> from tokenizers import ByteLevelBPETokenizer
    >>> from transformers.models.roberta import RobertaTokenizerFast
    >>> tokenizer = ByteLevelBPETokenizer()
    >>> tokenizer.train(files=filepath, vocab_size=1_000, min_frequency=2, special_tokens=["<s>", "<pad>", "</s>", "<unk>", "<mask>"])
    >>> tokenizer_path = os.path.join(tempdir, 'tokenizer')
    >>> os.makedirs(tokenizer_path)
    >>> result = tokenizer.save_model(tokenizer_path)
    >>> tokenizer = RobertaTokenizerFast.from_pretrained(tokenizer_path)

    >>> # preparing dataset
    >>> import pandas as pd
    >>> import deepchem as dc
    >>> smiles = ["CCN(CCSC)C(=O)N[C@@](C)(CC)C(F)(F)F","CC1(C)CN(C(=O)Nc2cc3ccccc3nn2)C[C@@]2(CCOC2)O1"]
    >>> labels = [3.112,2.432]
    >>> df = pd.DataFrame(list(zip(smiles, labels)), columns=["smiles", "task1"])
    >>> with dc.utils.UniversalNamedTemporaryFile(mode='w') as tmpfile:
    ...     df.to_csv(tmpfile.name)
    ...     loader = dc.data.CSVLoader(["task1"], feature_field="smiles", featurizer=dc.feat.DummyFeaturizer())
    ...     dataset = loader.create_dataset(tmpfile.name)

    >>> # pretraining
    >>> from deepchem.models.torch_models.hf_models import HuggingFaceModel
    >>> from transformers.models.roberta import RobertaForMaskedLM, RobertaModel, RobertaConfig
    >>> config = RobertaConfig(vocab_size=tokenizer.vocab_size)
    >>> model = RobertaForMaskedLM(config)
    >>> hf_model = HuggingFaceModel(model=model, tokenizer=tokenizer, task='mlm', model_dir='model-dir')
    >>> training_loss = hf_model.fit(dataset, nb_epoch=1)

    >>> # finetuning a regression model
    >>> from transformers.models.roberta import RobertaForSequenceClassification
    >>> config = RobertaConfig(vocab_size=tokenizer.vocab_size, problem_type='regression', num_labels=1)
    >>> model = RobertaForSequenceClassification(config)
    >>> hf_model = HuggingFaceModel(model=model, tokenizer=tokenizer, task='regression', model_dir='model-dir')
    >>> hf_model.load_from_pretrained()
    >>> training_loss = hf_model.fit(dataset, nb_epoch=1)
    >>> prediction = hf_model.predict(dataset)  # prediction
    >>> eval_results = hf_model.evaluate(dataset, metrics=dc.metrics.Metric(dc.metrics.mae_score))

    >>> # finetune a classification model
    >>> # making dataset suitable for classification
    >>> import numpy as np
    >>> y = np.random.choice([0, 1], size=dataset.y.shape)
    >>> dataset = dc.data.NumpyDataset(X=dataset.X, y=y, w=dataset.w, ids=dataset.ids)

    >>> from transformers import RobertaForSequenceClassification
    >>> config = RobertaConfig(vocab_size=tokenizer.vocab_size)
    >>> model = RobertaForSequenceClassification(config)
    >>> hf_model = HuggingFaceModel(model=model, task='classification', tokenizer=tokenizer)
    >>> training_loss = hf_model.fit(dataset, nb_epoch=1)
    >>> predictions = hf_model.predict(dataset)
    >>> eval_result = hf_model.evaluate(dataset, metrics=dc.metrics.Metric(dc.metrics.f1_score))
    """

    def __init__(
            self,
            model: 'PreTrainedModel',
            tokenizer: 'transformers.tokenization_utils.PreTrainedTokenizer',
            task: Optional[str] = None,
            **kwargs):
        self.task = task
        self.tokenizer = tokenizer
        if self.task == 'mlm':
            self.data_collator = DataCollatorForLanguageModeling(
                tokenizer=tokenizer)
        else:
            self.data_collator = None  # type: ignore
        # Ignoring type. For TorchModel, loss is a required argument but HuggingFace computes
        # loss during the forward iteration, removing the need for a loss function.
        super(HuggingFaceModel, self).__init__(
            model=model,
            loss=None,  # type: ignore
            **kwargs)

    def load_from_pretrained(  # type: ignore
            self,
            model_dir: Optional[str] = None,
            from_hf_checkpoint: bool = False):
        """Load HuggingFace model from a pretrained checkpoint.

        The utility can be used for loading a model from a checkpoint.
        Given `model_dir`, it checks for existing checkpoint in the directory.
        If a checkpoint exists, the models state is loaded from the checkpoint.

        If the option `from_hf_checkpoint` is set as True, then it loads a pretrained
        model using HuggingFace models `from_pretrained` method. This option
        interprets model_dir as a model id of a pretrained model hosted inside a model repo
        on huggingface.co or path to directory containing model weights saved using `save_pretrained`
        method of a HuggingFace model.

        Parameter
        ----------
        model_dir: str
            Directory containing model checkpoint
        from_hf_checkpoint: bool, default False
            Loads a pretrained model from HuggingFace checkpoint.

        Example
        -------
        >>> from transformers import RobertaTokenizerFast
        >>> tokenizer = RobertaTokenizerFast.from_pretrained("seyonec/PubChem10M_SMILES_BPE_60k")

        >>> from deepchem.models.torch_models.hf_models import HuggingFaceModel
        >>> from transformers.models.roberta import RobertaForMaskedLM, RobertaModel, RobertaConfig
        >>> config = RobertaConfig(vocab_size=tokenizer.vocab_size)
        >>> model = RobertaForMaskedLM(config)
        >>> pretrain_model = HuggingFaceModel(model=model, tokenizer=tokenizer, task='mlm', model_dir='model-dir')
        >>> pretrain_model.save_checkpoint()

        >>> from transformers import RobertaForSequenceClassification
        >>> config = RobertaConfig(vocab_size=tokenizer.vocab_size)
        >>> model = RobertaForSequenceClassification(config)
        >>> finetune_model = HuggingFaceModel(model=model, task='classification', tokenizer=tokenizer, model_dir='model-dir')

        >>> finetune_model.load_from_pretrained()
        """
        if model_dir is None:
            model_dir = self.model_dir

        if from_hf_checkpoint:
            # FIXME Transformers library has an api like AutoModel.from_pretrained. It allows to
            # initialise and create a model instance directly without requiring a class instance initialisation step.
            # To use `load_from_pretrained` in DeepChem, we need to follow a two step process
            # of initialising class instance and then loading weights via `load_from_pretrained`.
            if self.task == 'mlm':
                self.model = AutoModelForMaskedLM.from_pretrained(model_dir)
            elif self.task in ['mtr', 'regression', 'classification']:
                self.model = AutoModelForSequenceClassification.from_pretrained(
                    model_dir)
            else:
                self.model = AutoModel.from_pretrained(model_dir)
        elif not from_hf_checkpoint:
            checkpoints = sorted(self.get_checkpoints(model_dir))
            if len(checkpoints) == 0:
                raise ValueError('No checkpoint found')
            else:
                checkpoint = checkpoints[0]
                data = torch.load(checkpoint, map_location=self.device)
                self.model.load_state_dict(data['model_state_dict'],
                                           strict=False)

    def _prepare_batch(self, batch: Tuple[Any, Any, Any]):
        smiles_batch, y, w = batch
        tokens = self.tokenizer(smiles_batch[0].tolist(),
                                padding=True,
                                return_tensors="pt")

        if self.task == 'mlm':
            inputs, labels = self.data_collator.torch_mask_tokens(
                tokens['input_ids'])
            inputs = {
                'input_ids': inputs.to(self.device),
                'labels': labels.to(self.device),
                'attention_mask': tokens['attention_mask'].to(self.device),
            }
            return inputs, None, w
        elif self.task in ['regression', 'classification', 'mtr']:
            if y is not None:
                # y is None during predict
                y = torch.from_numpy(y[0])
                if self.task == 'regression' or self.task == 'mtr':
                    y = y.float().to(self.device)
                elif self.task == 'classification':
                    y = y.long().to(self.device)
            for key, value in tokens.items():
                tokens[key] = value.to(self.device)

            inputs = {**tokens, 'labels': y}
            return inputs, y, w

    def fit_generator(self,
                      generator: Iterable[Tuple[Any, Any, Any]],
                      max_checkpoints_to_keep: int = 5,
                      checkpoint_interval: int = 1000,
                      restore: bool = False,
                      variables: Optional[Union[List[torch.nn.Parameter],
                                                torch.nn.ParameterList]] = None,
                      loss: Optional[LossFn] = None,
                      callbacks: Union[Callable, List[Callable]] = [],
                      all_losses: Optional[List[float]] = None) -> float:
        """Train this model on data from a generator.

        Parameters
        ----------
        generator: generator
            this should generate batches, each represented as a tuple of the form
            (inputs, labels, weights).
        max_checkpoints_to_keep: int
            the maximum number of checkpoints to keep.  Older checkpoints are discarded.
        checkpoint_interval: int
            the frequency at which to write checkpoints, measured in training steps.
            Set this to 0 to disable automatic checkpointing.
        restore: bool
            if True, restore the model from the most recent checkpoint and continue training
            from there.  If False, retrain the model from scratch.
        variables: list of torch.nn.Parameter
            the variables to train.  If None (the default), all trainable variables in
            the model are used.
        loss: function
            a function of the form f(outputs, labels, weights) that computes the loss
            for each batch.  If None (the default), the model's standard loss function
            is used.
        callbacks: function or list of functions
            one or more functions of the form f(model, step) that will be invoked after
            every step.  This can be used to perform validation, logging, etc.
        all_losses: Optional[List[float]], optional (default None)
            If specified, all logged losses are appended into this list. Note that
            you can call `fit()` repeatedly with the same list and losses will
            continue to be appended.

        Returns
        -------
        The average loss over the most recent checkpoint interval

        Note
        ----
        A HuggingFace model can return embeddings (last hidden state), attentions.
        Support must be added to return the embeddings to the user, so that it can
        be used for other downstream applications.
        """
        if not isinstance(callbacks, SequenceCollection):
            callbacks = [callbacks]
        self._ensure_built()
        self.model.train()
        avg_loss = 0.0
        last_avg_loss = 0.0
        averaged_batches = 0
        if variables is None:
            optimizer = self._pytorch_optimizer
            lr_schedule = self._lr_schedule
        else:
            var_key = tuple(variables)
            if var_key in self._optimizer_for_vars:
                optimizer, lr_schedule = self._optimizer_for_vars[var_key]
            else:
                optimizer = self.optimizer._create_pytorch_optimizer(variables)
                if isinstance(self.optimizer.learning_rate,
                              LearningRateSchedule):
                    lr_schedule = self.optimizer.learning_rate._create_pytorch_schedule(
                        optimizer)
                else:
                    lr_schedule = None
                self._optimizer_for_vars[var_key] = (optimizer, lr_schedule)
        time1 = time.time()

        # Main training loop.

        for batch in generator:
            if restore:
                self.restore()
                restore = False
            inputs: OneOrMany[torch.Tensor]
            inputs, labels, weights = self._prepare_batch(batch)

            optimizer.zero_grad()
            outputs = self.model(**inputs)

            if self._loss_outputs is not None:
                outputs = [outputs[i] for i in self._loss_outputs]
            batch_loss = outputs.get("loss")
            batch_loss.backward()
            optimizer.step()
            if lr_schedule is not None:
                lr_schedule.step()
            self._global_step += 1
            current_step = self._global_step

            avg_loss += batch_loss

            # Report progress and write checkpoints.
            averaged_batches += 1
            should_log = (current_step % self.log_frequency == 0)
            if should_log:
                avg_loss = float(avg_loss) / averaged_batches
                logger.info('Ending global_step %d: Average loss %g' %
                            (current_step, avg_loss))
                if all_losses is not None:
                    all_losses.append(avg_loss)
                # Capture the last avg_loss in case of return since we're resetting to 0 now
                last_avg_loss = avg_loss
                avg_loss = 0.0
                averaged_batches = 0

            if checkpoint_interval > 0 and current_step % checkpoint_interval == checkpoint_interval - 1:
                self.save_checkpoint(max_checkpoints_to_keep)
            for c in callbacks:
                c(self, current_step)
            if self.tensorboard and should_log:
                self._log_scalar_to_tensorboard('loss', batch_loss,
                                                current_step)
            if (self.wandb_logger is not None) and should_log:
                all_data = dict({'train/loss': batch_loss})
                self.wandb_logger.log_data(all_data, step=current_step)

        # Report final results.
        if averaged_batches > 0:
            avg_loss = float(avg_loss) / averaged_batches
            logger.info('Ending global_step %d: Average loss %g' %
                        (current_step, avg_loss))
            if all_losses is not None:
                all_losses.append(avg_loss)
            last_avg_loss = avg_loss

        if checkpoint_interval > 0:
            self.save_checkpoint(max_checkpoints_to_keep)

        time2 = time.time()
        logger.info("TIMING: model fitting took %0.3f s" % (time2 - time1))
        return last_avg_loss

    def _predict(self, generator: Iterable[Tuple[Any, Any, Any]],
                 transformers: List[Transformer], uncertainty: bool,
                 other_output_types: Optional[OneOrMany[str]]):
        """Predicts output for data provided by generator.

        This is the private implementation of prediction. Do not
        call it directly. Instead call one of the public prediction methods.

        Parameters
        ----------
        generator: generator
            this should generate batches, each represented as a tuple of the form
            (inputs, labels, weights).
        transformers: list of dc.trans.Transformers
            Transformers that the input data has been transformed by.  The output
            is passed through these transformers to undo the transformations.
        uncertainty: bool
            specifies whether this is being called as part of estimating uncertainty.
            If True, it sets the training flag so that dropout will be enabled, and
            returns the values of the uncertainty outputs.
        other_output_types: list, optional
            Provides a list of other output_types (strings) to predict from model.

        Returns
        -------
            a NumPy array of the model produces a single output, or a list of arrays
            if it produces multiple outputs

        Note
        ----
        A HuggingFace model does not output uncertainity. The argument is here
        since it is also present in TorchModel. Similarly, other variables like
        other_output_types are also not used. Instead, a HuggingFace model outputs
        loss, logits, hidden state and attentions.
        """
        results: Optional[List[List[np.ndarray]]] = None
        variances: Optional[List[List[np.ndarray]]] = None
        if uncertainty and (other_output_types is not None):
            raise ValueError(
                'This model cannot compute uncertainties and other output types simultaneously. Please invoke one at a time.'
            )
        if uncertainty:
            if self._variance_outputs is None or len(
                    self._variance_outputs) == 0:
                raise ValueError('This model cannot compute uncertainties')
            if len(self._variance_outputs) != len(self._prediction_outputs):
                raise ValueError(
                    'The number of variances must exactly match the number of outputs'
                )
        if other_output_types:
            if self._other_outputs is None or len(self._other_outputs) == 0:
                raise ValueError(
                    'This model cannot compute other outputs since no other output_types were specified.'
                )
        self._ensure_built()
        self.model.eval()
        for batch in generator:
            inputs, labels, weights = batch
            inputs, _, _ = self._prepare_batch((inputs, None, None))

            # Invoke the model.
            output_values = self.model(**inputs)
            output_values = output_values.get('logits')

            if isinstance(output_values, torch.Tensor):
                output_values = [output_values]
            output_values = [t.detach().cpu().numpy() for t in output_values]
            # Apply tranformers and record results.
            if uncertainty:
                var = [output_values[i] for i in self._variance_outputs]
                if variances is None:
                    variances = [var]
                else:
                    for i, t in enumerate(var):
                        variances[i].append(t)
            access_values = []
            if other_output_types:
                access_values += self._other_outputs
            elif self._prediction_outputs is not None:
                access_values += self._prediction_outputs

            if len(access_values) > 0:
                output_values = [output_values[i] for i in access_values]

            if len(transformers) > 0:
                if len(output_values) > 1:
                    raise ValueError(
                        "predict() does not support Transformers for models with multiple outputs."
                    )
                elif len(output_values) == 1:
                    output_values = [
                        undo_transforms(output_values[0], transformers)
                    ]
            if results is None:
                results = [[] for i in range(len(output_values))]
            for i, t in enumerate(output_values):
                results[i].append(t)

        # Concatenate arrays to create the final results.
        final_results = []
        final_variances = []
        if results is not None:
            for r in results:
                final_results.append(np.concatenate(r, axis=0))

        if uncertainty and variances is not None:
            for v in variances:
                final_variances.append(np.concatenate(v, axis=0))
            return zip(final_results, final_variances)

        if len(final_results) == 1:
            return final_results[0]
        else:
            return np.array(final_results)



import torch
import torch.nn as nn
import torch.nn.functional as F

from deepchem.utils.typing import OneOrMany, ActivationFn
from deepchem.utils.pytorch_utils import get_activation

from collections.abc import Sequence as SequenceCollection
from typing import List, Tuple, Callable, Literal, Union


class ProgressiveMultitask(nn.Module):
    """Implements a progressive multitask neural network in PyTorch.

    Progressive networks allow for multitask learning where each task
    gets a new column of weights and lateral connections to previous tasks
    are added to the network. As a result, there is no exponential
    forgetting where previous tasks are ignored.

    Examples
    --------
    >>> import torch
    >>> import deepchem as dc
    >>> n_tasks = 4
    >>> n_features = 1024
    >>> n_classes = 2
    >>> sample = torch.randn(16, n_features)
    >>> model = dc.models.torch_models.ProgressiveMultitask(n_tasks=n_tasks, n_features=n_features, layer_sizes=[1024, 1024], mode="classification", n_classes=n_classes)
    >>> output = model(sample)
    >>> print(output[0].type(), output[1].type())
    torch.FloatTensor torch.FloatTensor
    >>> print(output[0].shape, output[1].shape)
    torch.Size([16, 4, 2]) torch.Size([16, 4, 2])

    References
    ----------
    See [1]_ for a full description of the progressive architecture

    .. [1] Rusu, Andrei A., et al. "Progressive neural networks." arXiv preprint
        arXiv:1606.04671 (2016).
    """

    def __init__(self,
                 n_tasks: int,
                 n_features: int,
                 layer_sizes: List[int] = [1000],
                 mode: Literal['regression', 'classification'] = 'regression',
                 alpha_init_stddevs: OneOrMany[float] = 0.02,
                 weight_init_stddevs: OneOrMany[float] = 0.02,
                 bias_init_consts: OneOrMany[float] = 1.0,
                 weight_decay_penalty: float = 0.0,
                 weight_decay_penalty_type: str = "l2",
                 activation_fns: OneOrMany[ActivationFn] = 'relu',
                 dropouts: OneOrMany[float] = 0.5,
                 n_classes: int = 1):
        """
        Parameters
        ----------
        n_tasks: int
            Number of tasks.
        n_features: int
            Size of input feature vector.
        layer_sizes: list of ints
            List of layer sizes.
        mode: str
            Type of model.  Must be 'regression' or 'classification'.
        alpha_init_stddevs: float or list of floats
            Standard deviation for truncated normal distribution to initialize
            alpha parameters.
        weight_init_stddevs: float or list of floats
            Standard deviation for truncated normal distribution to initialize
            weight parameters.
        bias_init_consts: float or list of floats
            Constant value to initialize bias parameters.
        weight_decay_penalty: float
            Amount of weight decay penalty to use.
        weight_decay_penalty_type: str
            Type of weight decay penalty.  Must be 'l1' or 'l2'.
        activation_fns: str or list of str
            Name of activation function(s) to use.
        dropouts: float or list of floats
            Dropout probability.
        n_classes: int
            The number of classes to predict per task.
        """
        if weight_decay_penalty != 0.0:
            raise ValueError("Weight decay is not currently supported")

        n_layers = len(layer_sizes)
        if not isinstance(weight_init_stddevs, SequenceCollection):
            weight_init_stddevs = [weight_init_stddevs] * n_layers
        if not isinstance(alpha_init_stddevs, SequenceCollection):
            alpha_init_stddevs = [alpha_init_stddevs] * n_layers
        if not isinstance(bias_init_consts, SequenceCollection):
            bias_init_consts = [bias_init_consts] * n_layers
        if not isinstance(dropouts, SequenceCollection):
            dropouts = [dropouts] * n_layers
        if isinstance(
                activation_fns,
                str) or not isinstance(activation_fns, SequenceCollection):
            activation_fns = [activation_fns] * n_layers

        self.mode = mode
        self.n_tasks: int = n_tasks
        self.n_features: int = n_features
        self.layer_sizes: List[int] = layer_sizes
        self.n_classes: int = n_classes
        self.weight_init_stddevs: SequenceCollection[
            float] = weight_init_stddevs
        self.alpha_init_stddevs: SequenceCollection[float] = alpha_init_stddevs
        self.bias_init_consts: SequenceCollection[float] = bias_init_consts
        self.dropouts: SequenceCollection[float] = dropouts
        self.activation_fns: SequenceCollection[Callable] = [
            get_activation(f) for f in activation_fns
        ]

        super(ProgressiveMultitask, self).__init__()

        self.layers: nn.ModuleList = nn.ModuleList()
        self.adapters: nn.ModuleList = nn.ModuleList()
        self.alphas: nn.ModuleList = nn.ModuleList()

        for task in range(n_tasks):
            layer_list = []
            adapter_list = []
            alpha_list = []
            prev_size = n_features
            for i, size in enumerate(self.layer_sizes):
                layer_list.append(self._init_linear(prev_size, size, i))

                if task > 0:
                    if i > 0:
                        adapter, alpha = self._get_adapter(
                            task, prev_size, size, i)
                        adapter_list.append(adapter)
                        alpha_list.append(alpha)

                prev_size = size

            layer_list.append(
                self._init_linear(prev_size, self.n_classes,
                                  len(self.layer_sizes)))
            self.layers.append(nn.ModuleList(layer_list))
            if task > 0:
                adapter, alpha = self._get_adapter(task, prev_size,
                                                   self.n_classes,
                                                   len(self.layer_sizes))
                adapter_list.append(adapter)
                alpha_list.append(alpha)

                self.adapters.append(nn.ModuleList(adapter_list))
                self.alphas.append(nn.ParameterList(alpha_list))

    def _get_adapter(self, task: int, prev_size: int, size: int,
                     layer_num: int) -> Tuple[nn.ModuleList, torch.Tensor]:
        """Creates the adapter layer between previous tasks and the current layer.

        Parameters
        ----------
        task: int
            Task number.
        prev_size: int
            Size of previous layer.
        size: int
            Size of current layer.
        layer_num: int
            Layer number.

        Returns
        -------
        adapter: nn.Sequential
            Adapter layer.
        alpha: torch.Tensor
            Alpha parameter.
        """
        adapter = nn.ModuleList([
            self._init_linear(prev_size * task, prev_size, layer_num),
            self._init_linear(prev_size, size, layer_num, use_bias=False),
        ])
        alpha_init_stddev = (self.alpha_init_stddevs[layer_num]
                             if layer_num < len(self.layer_sizes) else
                             self.alpha_init_stddevs[-1])
        alpha = torch.empty(1, requires_grad=True)
        nn.init.trunc_normal_(alpha, std=alpha_init_stddev)
        return adapter, alpha

    def _init_linear(self,
                     in_features: int,
                     out_features: int,
                     layer_num: int,
                     use_bias: bool = True) -> nn.Linear:
        """Initialises nn.Linear layer weight and bias parameters.

        Parameters
        ----------
        in_features: int
            Size of input feature vector.
        out_features: int
            Size of output feature vector.
        layer_num: int
            Layer number.
        use_bias: bool
            Whether to use bias for Linear layer. Default to True.

        Returns
        -------
        layer: nn.Linear
            Linear layer with initialised parameters.
        """
        if layer_num < len(self.layer_sizes):
            weight_init_stddev = self.weight_init_stddevs[layer_num]
            bias_init_const = self.bias_init_consts[layer_num]
        elif layer_num == len(self.layer_sizes):
            weight_init_stddev = self.weight_init_stddevs[-1]
            bias_init_const = self.bias_init_consts[-1]

        layer = nn.Linear(in_features, out_features, bias=use_bias)
        nn.init.trunc_normal_(layer.weight, std=weight_init_stddev)

        if use_bias:
            nn.init.constant_(layer.bias, bias_init_const)

        return layer

    def forward(
        self, x: torch.Tensor
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        """Forward pass through the network.

        Parameters
        ----------
        x: torch.Tensor
            Input tensor.

        Returns
        -------
        torch.Tensor
            The Model output tensor of shape (batch_size, n_tasks, n_outputs).

        * When self.mode = `regression`,
            It consists of the output of each task.
        * When self.mode = `classification`,
            It consists of the probability of each class for each task.

        torch.Tensor, optional
            This is only returned when self.mode = `classification`, the output consists of the
            logits for classes before softmax.
        """
        task_outputs: List[torch.Tensor] = []
        layer_logits: List[List[torch.Tensor]] = []
        for task in range(self.n_tasks):
            x_ = x
            layer_outputs = []
            for i, (layer, activation_fn, dropout) in enumerate(
                    zip(self.layers[task], self.activation_fns, self.dropouts)):
                x_ = layer(x_)

                if task > 0 and i > 0:
                    adapter_out = self._get_lateral_contrib(
                        layer_logits, task, i)
                    x_ = x_ + adapter_out

                if i < len(self.layer_sizes):
                    x_ = activation_fn(x_)
                    if dropout > 0.0:
                        x_ = F.dropout(x_, p=dropout, training=self.training)

                layer_outputs.append(x_)

            out_layer_idx = len(self.layer_sizes)
            x_ = self.layers[task][out_layer_idx](x_)
            if task > 0:
                adapter_out = self._get_lateral_contrib(layer_logits, task,
                                                        out_layer_idx)
                x_ = x_ + adapter_out

            layer_logits.append(layer_outputs)
            task_outputs.append(x_)

        output = torch.stack(task_outputs, dim=1)

        if self.mode == 'classification':
            if self.n_tasks == 1:
                logits = output.view(-1, self.n_classes)
                softmax_dim = 1
            else:
                logits = output.view(-1, self.n_tasks, self.n_classes)
                softmax_dim = 2
            proba = F.softmax(logits, dim=softmax_dim)
            return proba, logits
        else:
            return output

    def _get_lateral_contrib(self, logits, task, layer):
        adapter = self.adapters[task - 1][layer - 1]
        alpha = self.alphas[task - 1][layer - 1]
        prev_logits = [logits[t][layer - 1] for t in range(task)]
        adapter_input = torch.cat(prev_logits, dim=-1)
        adapter_input = alpha * adapter_input

        adapter_out = adapter[0](adapter_input)
        adapter_out = self.activation_fns[layer - 1](adapter_out)
        adapter_out = adapter[1](adapter_out)
        return adapter_out



"""
Contains basic hyperparameter optimizations.
"""
import numpy as np
import os
import itertools
import tempfile
import collections
import logging
from functools import reduce
from operator import mul
from typing import Dict, List, Optional, Tuple

from deepchem.data import Dataset
from deepchem.trans import Transformer
from deepchem.models import Model
from deepchem.metrics import Metric
from deepchem.hyper.base_classes import HyperparamOpt
from deepchem.hyper.base_classes import _convert_hyperparam_dict_to_filename

logger = logging.getLogger(__name__)


class GridHyperparamOpt(HyperparamOpt):
    """
    Provides simple grid hyperparameter search capabilities.

    This class performs a grid hyperparameter search over the specified
    hyperparameter space. This implementation is simple and simply does
    a direct iteration over all possible hyperparameters and doesn't use
    parallelization to speed up the search.

    Examples
    --------
    This example shows the type of constructor function expected.

    >>> import sklearn
    >>> import deepchem as dc
    >>> optimizer = dc.hyper.GridHyperparamOpt(lambda **p: dc.models.GraphConvModel(**p))

    Here's a more sophisticated example that shows how to optimize only
    some parameters of a model. In this case, we have some parameters we
    want to optimize, and others which we don't. To handle this type of
    search, we create a `model_builder` which hard codes some arguments
    (in this case, `max_iter` is a hyperparameter which we don't want
    to search over)

    >>> import deepchem as dc
    >>> import numpy as np
    >>> from sklearn.linear_model import LogisticRegression as LR
    >>> # generating data
    >>> X = np.arange(1, 11, 1).reshape(-1, 1)
    >>> y = np.hstack((np.zeros(5), np.ones(5)))
    >>> dataset = dc.data.NumpyDataset(X, y)
    >>> # splitting dataset into train and test
    >>> splitter = dc.splits.RandomSplitter()
    >>> train_dataset, test_dataset = splitter.train_test_split(dataset)
    >>> # metric to evaluate result of a set of parameters
    >>> metric = dc.metrics.Metric(dc.metrics.accuracy_score)
    >>> # defining `model_builder`
    >>> def model_builder(**model_params):
    ...   penalty = model_params['penalty']
    ...   solver = model_params['solver']
    ...   lr = LR(penalty=penalty, solver=solver, max_iter=100)
    ...   return dc.models.SklearnModel(lr)
    >>> # the parameters which are to be optimized
    >>> params = {
    ...   'penalty': ['l1', 'l2'],
    ...   'solver': ['liblinear', 'saga']
    ...   }
    >>> # Creating optimizer and searching over hyperparameters
    >>> optimizer = dc.hyper.GridHyperparamOpt(model_builder)
    >>> best_model, best_hyperparams, all_results = \
    optimizer.hyperparam_search(params, train_dataset, test_dataset, metric)
    >>> best_hyperparams  # the best hyperparameters
    {'penalty': 'l2', 'solver': 'saga'}

    """

    def hyperparam_search(
        self,
        params_dict: Dict,
        train_dataset: Dataset,
        valid_dataset: Dataset,
        metric: Metric,
        output_transformers: List[Transformer] = [],
        nb_epoch: int = 10,
        use_max: bool = True,
        logfile: str = 'results.txt',
        logdir: Optional[str] = None,
        **kwargs,
    ) -> Tuple[Model, Dict, Dict]:
        """Perform hyperparams search according to params_dict.

        Each key to hyperparams_dict is a model_param. The values should
        be a list of potential values for that hyperparam.

        Parameters
        ----------
        params_dict: Dict
            Maps hyperparameter names (strings) to lists of possible
            parameter values.
        train_dataset: Dataset
            dataset used for training
        valid_dataset: Dataset
            dataset used for validation(optimization on valid scores)
        metric: Metric
            metric used for evaluation
        output_transformers: list[Transformer]
            Transformers for evaluation. This argument is needed since
            `train_dataset` and `valid_dataset` may have been transformed
            for learning and need the transform to be inverted before
            the metric can be evaluated on a model.
        nb_epoch: int, (default 10)
            Specifies the number of training epochs during each iteration of optimization.
            Not used by all model types.
        use_max: bool, optional
            If True, return the model with the highest score. Else return
            model with the minimum score.
        logdir: str, optional
            The directory in which to store created models. If not set, will
            use a temporary directory.
        logfile: str, optional (default `results.txt`)
            Name of logfile to write results to. If specified, this is must
            be a valid file name. If not specified, results of hyperparameter
            search will be written to `logdir/results.txt`.

        Returns
        -------
        Tuple[`best_model`, `best_hyperparams`, `all_scores`]
        `(best_model, best_hyperparams, all_scores)` where `best_model` is
        an instance of `dc.model.Model`, `best_hyperparams` is a
        dictionary of parameters, and `all_scores` is a dictionary mapping
        string representations of hyperparameter sets to validation
        scores.

        Notes
        -----
        From DeepChem 2.6, the return type of `best_hyperparams` is a dictionary of
        parameters rather than a tuple of parameters as it was previously. The new
        changes have been made to standardize the behaviour across different
        hyperparameter optimization techniques available in DeepChem.
        """
        hyperparams = params_dict.keys()
        hyperparam_vals = params_dict.values()
        for hyperparam_list in params_dict.values():
            assert isinstance(hyperparam_list, collections.abc.Iterable)

        number_combinations = reduce(mul,
                                     [len(vals) for vals in hyperparam_vals])

        if use_max:
            best_validation_score = -np.inf
        else:
            best_validation_score = np.inf

        best_model = None
        all_scores = {}

        if logdir is not None:
            if not os.path.exists(logdir):
                os.makedirs(logdir, exist_ok=True)
            log_file = os.path.join(logdir, logfile)

        for ind, hyperparameter_tuple in enumerate(
                itertools.product(*hyperparam_vals)):
            model_params = {}
            logger.info("Fitting model %d/%d" % (ind + 1, number_combinations))
            # Construction dictionary mapping hyperparameter names to values
            hyper_params = dict(zip(hyperparams, hyperparameter_tuple))
            for hyperparam, hyperparam_val in zip(hyperparams,
                                                  hyperparameter_tuple):
                model_params[hyperparam] = hyperparam_val
            logger.info("hyperparameters: %s" % str(model_params))

            hp_str = _convert_hyperparam_dict_to_filename(hyper_params)
            if logdir is not None:
                model_dir = os.path.join(logdir, hp_str)
                logger.info("model_dir is %s" % model_dir)
                try:
                    os.makedirs(model_dir)
                except OSError:
                    if not os.path.isdir(model_dir):
                        logger.info(
                            "Error creating model_dir, using tempfile directory"
                        )
                        model_dir = tempfile.mkdtemp()
            else:
                model_dir = tempfile.mkdtemp()
            model_params['model_dir'] = model_dir
            model = self.model_builder(**model_params)
            # mypy test throws error, so ignoring it in try
            try:
                model.fit(train_dataset, nb_epoch=nb_epoch)  # type: ignore
            # Not all models have nb_epoch
            except TypeError:
                model.fit(train_dataset)
            try:
                model.save()
            # Some models autosave
            except NotImplementedError:
                pass

            multitask_scores = model.evaluate(valid_dataset, [metric],
                                              output_transformers)
            valid_score = multitask_scores[metric.name]
            all_scores[hp_str] = valid_score

            if (use_max and valid_score >= best_validation_score) or (
                    not use_max and valid_score <= best_validation_score):
                best_validation_score = valid_score
                best_hyperparams = hyper_params
                best_model = model

            logger.info(
                "Model %d/%d, Metric %s, Validation set %s: %f" %
                (ind + 1, number_combinations, metric.name, ind, valid_score))
            logger.info("\tbest_validation_score so far: %f" %
                        best_validation_score)
        if best_model is None:
            logger.info("No models trained correctly.")
            # arbitrarily return last model
            if logdir is not None:
                with open(log_file, 'w+') as f:
                    f.write(
                        "No model trained correctly. Arbitary models returned")
            best_model, best_hyperparams = model, hyperparameter_tuple  # type: ignore
            return best_model, best_hyperparams, all_scores
        multitask_scores = best_model.evaluate(train_dataset, [metric],
                                               output_transformers)
        train_score = multitask_scores[metric.name]
        logger.info("Best hyperparameters: %s" % str(best_hyperparams))
        logger.info("best train score: %f" % train_score)
        logger.info("best validation score: %f" % best_validation_score)
        if logdir is not None:
            with open(log_file, 'w+') as f:
                f.write("Best Hyperparameters dictionary %s\n" %
                        str(best_hyperparams))
                f.write("Best validation score %f\n" % best_validation_score)
                f.write("Best train_score: %f\n" % train_score)
        return best_model, best_hyperparams, all_scores



"""
Density Function Theory Configuration Utilities.
"""

from dataclasses import dataclass

__all__ = ["config"]


@dataclass
class _Config(object):
    """Contains the configuration for the DFT module

    Examples
    --------
    >>> from deepchem.utils.dft_utils.config import config
    >>> Memory_usage = 1024**4 # Sample Memory usage by some Object/Matrix
    >>> if Memory_usage > config.THRESHOLD_MEMORY :
    ...     print("Overload")
    Overload

    Attributes
    ----------
    THRESHOLD_MEMORY: int (default=10*1024**3)
        Threshold memory (matrix above this size should not be constructed)
    CHUNK_MEMORY: int (default=16*1024**2)
        The memory for splitting big tensors into chunks.
    VERBOSE: int (default=0)
        Allowed Verbosity level (Defines the level of detail)
        Used by Looger for maintaining Logs.

    Usage
    -----
    1. HamiltonCGTO: Usage it for splitting big tensors into chunks.

    """
    THRESHOLD_MEMORY: int = 10 * 1024**3  # in Bytes
    CHUNK_MEMORY: int = 16 * 1024**2  # in Bytes
    VERBOSE: int = 0


config = _Config()



"""Unit tests for evaluators."""
import pytest
import deepchem as dc
import numpy as np
import sklearn
from deepchem.utils.evaluate import Evaluator
from deepchem.utils.evaluate import GeneratorEvaluator
try:
    import tensorflow as tf  # noqa
    has_tensorflow = True
except:
    has_tensorflow = False

try:
    import torch  # noqa
    has_pytorch = True
except:
    has_pytorch = False


def test_multiclass_threshold_predictions():
    """Check prediction thresholding works correctly."""
    # Construct a random class probability matrix
    y = np.random.rand(10, 5)
    y_sums = np.sum(y, axis=1)
    y = y / y_sums[:, None]
    y_out = dc.metrics.threshold_predictions(y)
    assert y_out.shape == (10,)
    assert np.allclose(y_out, np.argmax(y, axis=1))


def test_binary_threshold_predictions():
    """Check prediction thresholding works correctly."""
    # Construct a random class probability matrix
    y = np.random.rand(10, 2)
    y_sums = np.sum(y, axis=1)
    y = y / y_sums[:, None]
    y_out = dc.metrics.threshold_predictions(y, threshold=0.3)
    assert y_out.shape == (10,)
    assert np.allclose(y_out, np.where(y[:, 1] >= 0.3, np.ones(10),
                                       np.zeros(10)))


@pytest.mark.torch
def test_evaluator_dc_metric():
    """Test an evaluator on a dataset."""
    X = np.random.rand(10, 5)
    y = np.random.rand(10, 1)
    dataset = dc.data.NumpyDataset(X, y)
    model = dc.models.MultitaskRegressor(1, 5)
    evaluator = Evaluator(model, dataset, [])
    metric = dc.metrics.Metric(dc.metrics.mae_score)
    multitask_scores = evaluator.compute_model_performance(metric)
    assert isinstance(multitask_scores, dict)
    assert len(multitask_scores) == 1
    assert multitask_scores['mae_score'] > 0


@pytest.mark.torch
def test_multiclass_classification_singletask():
    """Test multiclass classification evaluation."""
    X = np.random.rand(100, 5)
    y = np.random.randint(5, size=(100,))
    dataset = dc.data.NumpyDataset(X, y)
    model = dc.models.MultitaskClassifier(1, 5, n_classes=5)
    evaluator = Evaluator(model, dataset, [])
    multitask_scores = evaluator.compute_model_performance(
        dc.metrics.roc_auc_score, n_classes=5)
    assert len(multitask_scores) == 1
    assert multitask_scores["metric-1"] >= 0


def test_sklearn_multiclass_classification_singletask():
    """Test multiclass classification evaluation."""
    X = np.random.rand(100, 5)
    y = np.random.randint(5, size=(100,))
    dataset = dc.data.NumpyDataset(X, y)
    rf = sklearn.ensemble.RandomForestClassifier(50)
    model = dc.models.SklearnModel(rf)
    model.fit(dataset)
    evaluator = Evaluator(model, dataset, [])
    multitask_scores = evaluator.compute_model_performance(
        dc.metrics.roc_auc_score, n_classes=5)
    assert len(multitask_scores) == 1
    assert multitask_scores["metric-1"] >= 0


@pytest.mark.torch
def test_evaluate_multiclass_classification_singletask():
    """Test multiclass classification evaluation."""
    X = np.random.rand(100, 5)
    y = np.random.randint(5, size=(100,))
    dataset = dc.data.NumpyDataset(X, y)
    model = dc.models.MultitaskClassifier(1, 5, n_classes=5)
    multitask_scores = model.evaluate(dataset,
                                      dc.metrics.roc_auc_score,
                                      n_classes=5)
    assert len(multitask_scores) == 1
    assert multitask_scores["metric-1"] >= 0


@pytest.mark.torch
def test_multitask_evaluator():
    """Test evaluation of a multitask metric."""
    X = np.random.rand(10, 5)
    y = np.random.rand(10, 2, 1)
    dataset = dc.data.NumpyDataset(X, y)
    model = dc.models.MultitaskRegressor(2, 5)
    evaluator = Evaluator(model, dataset, [])
    metric = dc.metrics.Metric(dc.metrics.mae_score)
    multitask_scores, all_task_scores = evaluator.compute_model_performance(
        metric, per_task_metrics=True)
    assert isinstance(multitask_scores, dict)
    assert len(multitask_scores) == 1
    assert multitask_scores['mae_score'] > 0
    assert isinstance(all_task_scores, dict)
    assert len(multitask_scores) == 1


@pytest.mark.torch
def test_model_evaluate_dc_metric():
    """Test a model evaluate on a dataset."""
    X = np.random.rand(10, 5)
    y = np.random.rand(10, 1)
    dataset = dc.data.NumpyDataset(X, y)
    model = dc.models.MultitaskRegressor(1, 5)
    metric = dc.metrics.Metric(dc.metrics.mae_score)
    multitask_scores = model.evaluate(dataset, metric, [])
    assert isinstance(multitask_scores, dict)
    assert len(multitask_scores) == 1
    assert multitask_scores['mae_score'] > 0


@pytest.mark.torch
def test_multitask_model_evaluate_sklearn():
    """Test evaluation of a multitask metric."""
    X = np.random.rand(10, 5)
    y = np.random.rand(10, 2)
    dataset = dc.data.NumpyDataset(X, y)
    model = dc.models.MultitaskRegressor(2, 5)
    evaluator = Evaluator(model, dataset, [])
    multitask_scores, all_task_scores = evaluator.compute_model_performance(
        dc.metrics.mean_absolute_error, per_task_metrics=True)
    assert isinstance(multitask_scores, dict)
    assert len(multitask_scores) == 1
    assert multitask_scores['metric-1'] > 0
    assert isinstance(all_task_scores, dict)
    assert len(multitask_scores) == 1


@pytest.mark.torch
def test_multitask_model_evaluate():
    """Test evaluation of a multitask metric."""
    X = np.random.rand(10, 5)
    y = np.random.rand(10, 2)
    dataset = dc.data.NumpyDataset(X, y)
    model = dc.models.MultitaskRegressor(2, 5)
    multitask_scores, all_task_scores = model.evaluate(
        dataset, dc.metrics.mean_absolute_error, per_task_metrics=True)
    assert isinstance(multitask_scores, dict)
    assert len(multitask_scores) == 1
    assert multitask_scores["metric-1"] > 0
    assert isinstance(all_task_scores, dict)


@pytest.mark.torch
def test_evaluator_dc_multi_metric():
    """Test an evaluator on a dataset."""
    X = np.random.rand(10, 5)
    y = np.random.rand(10, 1)
    dataset = dc.data.NumpyDataset(X, y)
    model = dc.models.MultitaskRegressor(1, 5)
    evaluator = Evaluator(model, dataset, [])
    metric1 = dc.metrics.Metric(dc.metrics.mae_score, n_tasks=2)
    metric2 = dc.metrics.Metric(dc.metrics.r2_score, n_tasks=2)
    multitask_scores = evaluator.compute_model_performance([metric1, metric2])
    assert isinstance(multitask_scores, dict)
    assert len(multitask_scores) == 2
    assert multitask_scores['mae_score'] > 0
    assert "r2_score" in multitask_scores


@pytest.mark.torch
def test_model_evaluate_dc_multi_metric():
    """Test an evaluator on a dataset."""
    X = np.random.rand(10, 5)
    y = np.random.rand(10, 1)
    dataset = dc.data.NumpyDataset(X, y)
    model = dc.models.MultitaskRegressor(1, 5)
    metric1 = dc.metrics.Metric(dc.metrics.mae_score)
    metric2 = dc.metrics.Metric(dc.metrics.r2_score)
    multitask_scores = model.evaluate(dataset, [metric1, metric2])
    assert isinstance(multitask_scores, dict)
    assert len(multitask_scores) == 2
    assert multitask_scores['mae_score'] > 0
    assert "r2_score" in multitask_scores


@pytest.mark.torch
def test_generator_evaluator_dc_metric_multitask_single_point():
    """Test generator evaluator on a generator."""
    X = np.random.rand(10, 5)
    y = np.random.rand(10, 1)
    dataset = dc.data.NumpyDataset(X, y)
    model = dc.models.MultitaskRegressor(1, 5)
    generator = model.default_generator(dataset, pad_batches=False)
    evaluator = GeneratorEvaluator(model, generator, [])
    metric = dc.metrics.Metric(dc.metrics.mae_score)
    multitask_scores = evaluator.compute_model_performance(metric)
    assert isinstance(multitask_scores, dict)
    assert len(multitask_scores) == 1
    assert multitask_scores['mae_score'] > 0
    assert len(multitask_scores) == 1


@pytest.mark.torch
def test_evaluator_sklearn_metric():
    """Test an evaluator on a dataset."""
    X = np.random.rand(10, 5)
    y = np.random.rand(10, 1)
    dataset = dc.data.NumpyDataset(X, y)
    model = dc.models.MultitaskRegressor(1, 5)
    evaluator = Evaluator(model, dataset, [])
    multitask_scores = evaluator.compute_model_performance(
        dc.metrics.mean_absolute_error)
    assert isinstance(multitask_scores, dict)
    assert len(multitask_scores) == 1
    # Note that since no name as provided, metrics are index by order
    # given.
    assert multitask_scores['metric-1'] > 0


@pytest.mark.torch
def test_generator_evaluator_dc_metric_multitask():
    """Test generator evaluator on a generator."""
    X = np.random.rand(10, 5)
    y = np.random.rand(10, 1)
    dataset = dc.data.NumpyDataset(X, y)
    model = dc.models.MultitaskRegressor(1, 5)
    generator = model.default_generator(dataset, pad_batches=False)
    evaluator = GeneratorEvaluator(model, generator, [])
    metric = dc.metrics.Metric(dc.metrics.mae_score)
    multitask_scores = evaluator.compute_model_performance(metric)
    assert isinstance(multitask_scores, dict)
    assert len(multitask_scores) == 1
    assert multitask_scores['mae_score'] > 0


@pytest.mark.torch
def test_model_evaluate_sklearn_metric():
    """Test a model evaluate on a dataset."""
    X = np.random.rand(10, 5)
    y = np.random.rand(10, 1)
    dataset = dc.data.NumpyDataset(X, y)
    model = dc.models.MultitaskRegressor(1, 5)
    multitask_scores = model.evaluate(dataset, dc.metrics.mean_absolute_error)
    assert isinstance(multitask_scores, dict)
    assert len(multitask_scores) == 1
    # Note that since no name as provided, metrics are index by order
    # given.
    assert multitask_scores['metric-1'] > 0


@pytest.mark.torch
def test_evaluator_sklearn_multi_metric():
    """Test an evaluator on a dataset."""
    X = np.random.rand(10, 5)
    y = np.random.rand(10, 1)
    dataset = dc.data.NumpyDataset(X, y)
    model = dc.models.MultitaskRegressor(1, 5)
    evaluator = Evaluator(model, dataset, [])
    multitask_scores = evaluator.compute_model_performance(
        [dc.metrics.mean_absolute_error, dc.metrics.r2_score])
    assert isinstance(multitask_scores, dict)
    assert len(multitask_scores.keys()) == 2
    # Note that since no name as provided, metrics are index by order
    # given.
    assert multitask_scores['metric-1'] > 0
    assert "metric-2" in multitask_scores


@pytest.mark.torch
def test_model_evaluate_sklearn_multi_metric():
    """Test an evaluator on a dataset."""
    X = np.random.rand(10, 5)
    y = np.random.rand(10, 1)
    dataset = dc.data.NumpyDataset(X, y)
    model = dc.models.MultitaskRegressor(1, 5)
    multitask_scores = model.evaluate(
        dataset, [dc.metrics.mean_absolute_error, dc.metrics.r2_score])
    assert isinstance(multitask_scores, dict)
    assert len(multitask_scores.keys()) == 2
    # Note that since no name as provided, metrics are index by order
    # given.
    assert multitask_scores['metric-1'] > 0
    assert "metric-2" in multitask_scores


@pytest.mark.tensorflow
def test_gc_binary_classification():
    """Test multiclass classification evaluation."""
    smiles = ["C", "CC"]
    featurizer = dc.feat.ConvMolFeaturizer()
    X = featurizer.featurize(smiles)
    y = np.random.randint(2, size=(len(smiles),))
    dataset = dc.data.NumpyDataset(X, y)
    model = dc.models.GraphConvModel(1, mode="classification")
    # TODO: Fix this case with correct thresholding
    evaluator = Evaluator(model, dataset, [])
    multitask_scores = evaluator.compute_model_performance(
        dc.metrics.accuracy_score, n_classes=2)
    assert len(multitask_scores) == 1
    assert multitask_scores["metric-1"] >= 0


@pytest.mark.tensorflow
def test_gc_binary_kappa_classification():
    """Test multiclass classification evaluation."""
    np.random.seed(1234)
    smiles = ["C", "CC", "CO", "CCC", "CCCC"]
    featurizer = dc.feat.ConvMolFeaturizer()
    X = featurizer.featurize(smiles)
    y = np.random.randint(2, size=(len(smiles),))
    dataset = dc.data.NumpyDataset(X, y)
    model = dc.models.GraphConvModel(1, mode="classification")
    # TODO: Fix this case with correct thresholding
    evaluator = Evaluator(model, dataset, [])
    multitask_scores = evaluator.compute_model_performance(
        dc.metrics.kappa_score, n_classes=2)
    assert len(multitask_scores) == 1
    assert multitask_scores["metric-1"] <= 1
    assert multitask_scores["metric-1"] >= -1


@pytest.mark.tensorflow
def test_gc_multiclass_classification():
    """Test multiclass classification evaluation."""
    np.random.seed(1234)
    smiles = ["C", "CC"]
    featurizer = dc.feat.ConvMolFeaturizer()
    X = featurizer.featurize(smiles)
    y = np.random.randint(5, size=(len(smiles),))
    dataset = dc.data.NumpyDataset(X, y)
    model = dc.models.GraphConvModel(1, mode="classification", n_classes=5)
    evaluator = Evaluator(model, dataset, [])
    multitask_scores = evaluator.compute_model_performance(
        dc.metrics.accuracy_score, n_classes=5)
    assert len(multitask_scores) == 1
    assert multitask_scores["metric-1"] >= 0
